"""Tests for git context capture."""

import os
import subprocess
from typing import Any, cast

import pytest

import llmdebug.git_context as git_context_module
from llmdebug.git_context import _GitError, get_enhanced_git_context, get_git_context


@pytest.fixture(autouse=True)
def _clear_git_context_cache():
    """Prevent cross-test leakage from git context cache."""
    git_context_module._git_context_cache = None
    yield
    git_context_module._git_context_cache = None


def test_git_context_in_repo():
    """Test git context capture when in a git repository."""
    # This test runs from the llmdebug repo itself
    context = get_git_context()

    # Should have context since we're in a git repo
    assert context is not None
    assert "commit" in context
    assert "commit_full" in context
    assert "branch" in context
    assert "dirty" in context

    # Commit should be a short hash (7+ chars)
    assert len(context["commit"]) >= 7
    # Full commit should be 40 chars
    assert len(context["commit_full"]) == 40


def test_git_context_not_in_repo(tmp_path, monkeypatch):
    """Test git context capture when not in a git repository."""
    # Change to a non-git directory
    monkeypatch.chdir(tmp_path)

    context = get_git_context()

    # Should return None when not in a git repo
    assert context is None


def test_git_context_dirty_status(tmp_path):
    """Test that dirty status is correctly detected."""
    # Initialize a new git repo
    subprocess.run(["git", "init"], cwd=tmp_path, capture_output=True, check=True)
    subprocess.run(
        ["git", "config", "user.email", "test@test.com"],
        cwd=tmp_path,
        capture_output=True,
        check=True,
    )
    subprocess.run(
        ["git", "config", "user.name", "Test"],
        cwd=tmp_path,
        capture_output=True,
        check=True,
    )

    # Create and commit a file
    test_file = tmp_path / "test.txt"
    test_file.write_text("content")
    subprocess.run(["git", "add", "."], cwd=tmp_path, capture_output=True, check=True)
    subprocess.run(
        ["git", "commit", "-m", "Initial commit"],
        cwd=tmp_path,
        capture_output=True,
        check=True,
    )

    # Save current dir and change to temp repo
    old_cwd = os.getcwd()
    try:
        os.chdir(tmp_path)

        # Should be clean
        context = get_git_context()
        assert context is not None
        assert context["dirty"] is False

        # Untracked files should count as dirty.
        untracked_file = tmp_path / "untracked.txt"
        untracked_file.write_text("new")
        context = get_git_context()
        assert context is not None
        assert context["dirty"] is True

        # Return to clean before testing tracked-file edits.
        untracked_file.unlink()
        context = get_git_context()
        assert context is not None
        assert context["dirty"] is False

        # Tracked-file edits should also count as dirty.
        test_file.write_text("modified")
        context = get_git_context()
        assert context is not None
        assert context["dirty"] is True
    finally:
        os.chdir(old_cwd)


def test_git_context_detached_head(tmp_path):
    """Test git context when HEAD is detached."""
    # Initialize a new git repo
    subprocess.run(["git", "init"], cwd=tmp_path, capture_output=True, check=True)
    subprocess.run(
        ["git", "config", "user.email", "test@test.com"],
        cwd=tmp_path,
        capture_output=True,
        check=True,
    )
    subprocess.run(
        ["git", "config", "user.name", "Test"],
        cwd=tmp_path,
        capture_output=True,
        check=True,
    )

    # Create and commit a file
    test_file = tmp_path / "test.txt"
    test_file.write_text("content")
    subprocess.run(["git", "add", "."], cwd=tmp_path, capture_output=True, check=True)
    subprocess.run(
        ["git", "commit", "-m", "Initial commit"],
        cwd=tmp_path,
        capture_output=True,
        check=True,
    )

    # Detach HEAD
    result = subprocess.run(
        ["git", "rev-parse", "HEAD"],
        cwd=tmp_path,
        capture_output=True,
        text=True,
        check=True,
    )
    commit_hash = result.stdout.strip()
    subprocess.run(
        ["git", "checkout", commit_hash],
        cwd=tmp_path,
        capture_output=True,
        check=True,
    )

    # Save current dir and change to temp repo
    old_cwd = os.getcwd()
    try:
        os.chdir(tmp_path)

        context = get_git_context()
        assert context is not None
        assert context.get("detached") is True
    finally:
        os.chdir(old_cwd)


def test_enhanced_git_context_parses_metadata(monkeypatch):
    snapshot = {
        "crash_frame_index": 0,
        "frames": [{"file_rel": "app.py", "line": 42}],
    }

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = timeout
        _ = repo_root
        if args[:2] == ["rev-parse", "--show-toplevel"]:
            return "/repo"
        if args[:3] == ["rev-parse", "--short", "HEAD"]:
            return "abc1234"
        if args[:2] == ["rev-parse", "HEAD"]:
            return "abc1234567890abcdef1234567890abcdef1234"
        if args[:3] == ["rev-parse", "--abbrev-ref", "HEAD"]:
            return "main"
        if args[:2] == ["status", "--porcelain"]:
            return ""
        if args and args[0] == "blame":
            return (
                "def56789abcdef0123456789abcdef0123456789 42 42 1\n"
                "author Alice\n"
                "author-time 1700000000\n"
                "summary Fix parser line\n"
                "filename app.py\n"
                "\treturn int(value)\n"
            )
        if args and args[0] == "log":
            return (
                "__LLMDEBUG_COMMIT__\n"
                "abc1234\n"
                "abc1234567890abcdef1234567890abcdef1234\n"
                "Alice\n"
                "2026-02-10T10:00:00+00:00\n"
                "Fix parser\n"
                " 1 file changed, 2 insertions(+), 1 deletion(-)\n"
            )
        if args and args[0] == "show":
            return "12\t3\tapp.py\n"
        return _GitError("unexpected")

    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    context = get_enhanced_git_context(
        snapshot,
        out_dir=".llmdebug",
        max_commits=5,
        per_cmd_timeout_sec=1.0,
        total_budget_sec=5.0,
    )
    assert context is not None
    assert context["head"]["commit"] == "abc1234"
    assert context["crash_file"]["file_rel"] == "app.py"
    assert context["crash_file"]["blame"]["author"] == "Alice"
    assert len(context["recent_commits"]) == 1
    assert context["crash_file_diffstat"]["files_changed"] == 1


def test_enhanced_git_context_discovers_repo_from_out_dir(monkeypatch, tmp_path):
    target_repo = (tmp_path / "target_repo").resolve()
    target_repo_str = target_repo.as_posix()
    target_out_dir = (target_repo / ".llmdebug").resolve()
    target_out_dir_str = target_out_dir.as_posix()
    wrong_repo_str = (tmp_path / "wrong_repo").resolve().as_posix()
    snapshot = {
        "crash_frame_index": 0,
        "frames": [{"file_rel": "app.py"}],
    }

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = timeout
        repo = str(repo_root) if repo_root is not None else None
        if args[:2] == ["rev-parse", "--show-toplevel"]:
            if repo in {target_out_dir_str, target_repo_str}:
                return target_repo_str
            return wrong_repo_str
        if args[:3] == ["rev-parse", "--short", "HEAD"]:
            return "abc1234" if repo == target_repo_str else "badcafe"
        if args[:2] == ["rev-parse", "HEAD"]:
            return ("a" * 40) if repo == target_repo_str else ("b" * 40)
        if args[:3] == ["rev-parse", "--abbrev-ref", "HEAD"]:
            return "main" if repo == target_repo_str else "wrong-branch"
        if args[:2] == ["status", "--porcelain"]:
            return ""
        if args and args[0] == "log":
            return ""
        return _GitError("unexpected")

    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    context = get_enhanced_git_context(snapshot, out_dir=target_out_dir_str)
    assert context is not None
    assert context["repo_root"] == target_repo_str
    assert context["head"]["commit"] == "abc1234"
    assert context["head"]["branch"] == "main"


def test_enhanced_git_context_marks_outside_repo(monkeypatch):
    snapshot = {
        "crash_frame_index": 0,
        "frames": [{"file": "/tmp/outside.py", "line": 8}],
    }
    calls: list[tuple[list[str], str | None]] = []

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = timeout
        calls.append((args, repo_root))
        if args[:2] == ["rev-parse", "--show-toplevel"]:
            return "/repo"
        if args[:3] == ["rev-parse", "--short", "HEAD"]:
            return "abc1234"
        if args[:2] == ["rev-parse", "HEAD"]:
            return "abc1234567890abcdef1234567890abcdef1234"
        if args[:3] == ["rev-parse", "--abbrev-ref", "HEAD"]:
            return "main"
        if args[:2] == ["status", "--porcelain"]:
            return ""
        if args and args[0] == "log":
            return ""
        return _GitError("unexpected")

    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    context = get_enhanced_git_context(snapshot, out_dir=".llmdebug")
    assert context is not None
    assert context["crash_file"]["in_repo"] is False
    assert "blame" not in context["crash_file"]
    assert all(args[0] not in {"blame", "show"} for args, _ in calls)


def test_enhanced_git_context_budget_exhausted(monkeypatch):
    snapshot = {"frames": [{"file_rel": "app.py", "line": 1}], "crash_frame_index": 0}

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = args
        _ = timeout
        _ = repo_root
        return "/repo"

    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    context = get_enhanced_git_context(snapshot, out_dir=".llmdebug", total_budget_sec=0.0)
    assert context is None


def test_git_context_cache_reuses_static_data(monkeypatch, tmp_path):
    repo_root = tmp_path / "repo"
    git_dir = repo_root / ".git"
    git_dir.mkdir(parents=True)
    (git_dir / "HEAD").write_text("ref: refs/heads/main\n", encoding="utf-8")
    (git_dir / "index").write_bytes(b"")
    monkeypatch.chdir(repo_root)
    monkeypatch.setattr(git_context_module, "_GIT_CONTEXT_CACHE_TTL_SEC", 60.0)

    run_calls: list[tuple[list[str], str | None]] = []
    rc_calls: list[tuple[list[str], str | None]] = []

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = timeout
        run_calls.append((args, repo_root))
        if args[:2] == ["rev-parse", "--show-toplevel"]:
            return str(repo_root_path)
        if args[:3] == ["rev-parse", "--short", "HEAD"]:
            return "abc1234"
        if args[:2] == ["rev-parse", "HEAD"]:
            return "a" * 40
        if args[:3] == ["rev-parse", "--abbrev-ref", "HEAD"]:
            return "main"
        return _GitError("unexpected")

    def fake_rc(args, timeout=2.0, repo_root=None):
        _ = timeout
        rc_calls.append((args, repo_root))
        return 0

    repo_root_path = repo_root.resolve()
    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    monkeypatch.setattr(git_context_module, "_run_git_command_rc", fake_rc)

    first = get_git_context()
    second = get_git_context()

    assert first is not None
    assert second is not None
    assert first["commit"] == "abc1234"
    assert second["commit"] == "abc1234"

    # Static metadata should be fetched once (cache hit on second call).
    static_calls = [args for args, _ in run_calls if args[:2] == ["rev-parse", "--show-toplevel"]]
    assert len(static_calls) == 1

    # Dirty checks still run on each access to avoid stale status.
    diff_calls = [args for args, _ in rc_calls if args[:1] == ["diff-files"]]
    assert len(diff_calls) == 2


def test_git_context_cache_invalidates_on_head_change(monkeypatch, tmp_path):
    repo_root = tmp_path / "repo"
    git_dir = repo_root / ".git"
    git_dir.mkdir(parents=True)
    head_path = git_dir / "HEAD"
    head_path.write_text("ref: refs/heads/main\n", encoding="utf-8")
    (git_dir / "index").write_bytes(b"")
    monkeypatch.chdir(repo_root)
    monkeypatch.setattr(git_context_module, "_GIT_CONTEXT_CACHE_TTL_SEC", 60.0)

    commit_state = {"short": "aaa1111", "full": "a" * 40}
    show_toplevel_calls = 0

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = timeout
        nonlocal show_toplevel_calls
        if args[:2] == ["rev-parse", "--show-toplevel"]:
            show_toplevel_calls += 1
            return str(repo_root_path)
        if args[:3] == ["rev-parse", "--short", "HEAD"]:
            return commit_state["short"]
        if args[:2] == ["rev-parse", "HEAD"]:
            return commit_state["full"]
        if args[:3] == ["rev-parse", "--abbrev-ref", "HEAD"]:
            return "main"
        return _GitError("unexpected")

    def fake_rc(args, timeout=2.0, repo_root=None):
        _ = args
        _ = timeout
        _ = repo_root
        return 0

    repo_root_path = repo_root.resolve()
    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    monkeypatch.setattr(git_context_module, "_run_git_command_rc", fake_rc)

    first = get_git_context()
    assert first is not None
    assert first["commit"] == "aaa1111"
    assert show_toplevel_calls == 1

    commit_state["short"] = "bbb2222"
    commit_state["full"] = "b" * 40
    # Head mtime change should invalidate cache and trigger refresh.
    head_path.write_text("ref: refs/heads/main\n# changed\n", encoding="utf-8")
    second = get_git_context()
    assert second is not None
    assert second["commit"] == "bbb2222"
    assert show_toplevel_calls == 2


def test_git_context_dirty_mode_off_skips_dirty_detection(monkeypatch, tmp_path):
    repo_root = tmp_path / "repo"
    git_dir = repo_root / ".git"
    git_dir.mkdir(parents=True)
    (git_dir / "HEAD").write_text("ref: refs/heads/main\n", encoding="utf-8")
    (git_dir / "index").write_bytes(b"")
    monkeypatch.chdir(repo_root)
    monkeypatch.setattr(git_context_module, "_GIT_CONTEXT_CACHE_TTL_SEC", 60.0)

    run_calls: list[list[str]] = []
    rc_calls: list[list[str]] = []

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = timeout
        _ = repo_root
        run_calls.append(args)
        if args[:2] == ["rev-parse", "--show-toplevel"]:
            return str(repo_root_path)
        if args[:3] == ["rev-parse", "--short", "HEAD"]:
            return "abc1234"
        if args[:2] == ["rev-parse", "HEAD"]:
            return "a" * 40
        if args[:3] == ["rev-parse", "--abbrev-ref", "HEAD"]:
            return "main"
        return _GitError("unexpected")

    def fake_rc(args, timeout=2.0, repo_root=None):
        _ = timeout
        _ = repo_root
        rc_calls.append(args)
        return 0

    repo_root_path = repo_root.resolve()
    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    monkeypatch.setattr(git_context_module, "_run_git_command_rc", fake_rc)

    context = get_git_context(dirty_mode="off")
    assert context is not None
    assert "dirty" not in context
    assert rc_calls == []
    assert all(call[:2] != ["status", "--porcelain"] for call in run_calls)


def test_git_context_dirty_mode_cached_reuses_dirty(monkeypatch, tmp_path):
    repo_root = tmp_path / "repo"
    git_dir = repo_root / ".git"
    git_dir.mkdir(parents=True)
    (git_dir / "HEAD").write_text("ref: refs/heads/main\n", encoding="utf-8")
    (git_dir / "index").write_bytes(b"")
    monkeypatch.chdir(repo_root)
    monkeypatch.setattr(git_context_module, "_GIT_CONTEXT_CACHE_TTL_SEC", 60.0)

    rc_calls: list[list[str]] = []

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = timeout
        _ = repo_root
        if args[:2] == ["rev-parse", "--show-toplevel"]:
            return str(repo_root_path)
        if args[:3] == ["rev-parse", "--short", "HEAD"]:
            return "abc1234"
        if args[:2] == ["rev-parse", "HEAD"]:
            return "a" * 40
        if args[:3] == ["rev-parse", "--abbrev-ref", "HEAD"]:
            return "main"
        return _GitError("unexpected")

    def fake_rc(args, timeout=2.0, repo_root=None):
        _ = timeout
        _ = repo_root
        rc_calls.append(args)
        # Mark repo dirty on first detection path.
        return 1

    repo_root_path = repo_root.resolve()
    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    monkeypatch.setattr(git_context_module, "_run_git_command_rc", fake_rc)

    first = get_git_context(dirty_mode="cached")
    second = get_git_context(dirty_mode="cached")
    assert first is not None
    assert second is not None
    assert first.get("dirty") is True
    assert second.get("dirty") is True

    diff_files_calls = [call for call in rc_calls if call[:1] == ["diff-files"]]
    assert len(diff_files_calls) == 1


def test_git_context_dirty_mode_cached_after_off_recomputes_dirty(monkeypatch, tmp_path):
    repo_root = tmp_path / "repo"
    git_dir = repo_root / ".git"
    git_dir.mkdir(parents=True)
    (git_dir / "HEAD").write_text("ref: refs/heads/main\n", encoding="utf-8")
    (git_dir / "index").write_bytes(b"")
    monkeypatch.chdir(repo_root)
    monkeypatch.setattr(git_context_module, "_GIT_CONTEXT_CACHE_TTL_SEC", 60.0)

    run_calls: list[list[str]] = []
    rc_calls: list[list[str]] = []

    def fake_run(args, timeout=2.0, repo_root=None):
        _ = timeout
        _ = repo_root
        run_calls.append(args)
        if args[:2] == ["rev-parse", "--show-toplevel"]:
            return str(repo_root_path)
        if args[:3] == ["rev-parse", "--short", "HEAD"]:
            return "abc1234"
        if args[:2] == ["rev-parse", "HEAD"]:
            return "a" * 40
        if args[:3] == ["rev-parse", "--abbrev-ref", "HEAD"]:
            return "main"
        return _GitError("unexpected")

    def fake_rc(args, timeout=2.0, repo_root=None):
        _ = timeout
        _ = repo_root
        rc_calls.append(args)
        if args[:1] == ["diff-files"]:
            return 0
        if args[:1] == ["diff-index"]:
            return 0
        if args[:1] == ["ls-files"]:
            return 1
        return 0

    repo_root_path = repo_root.resolve()
    monkeypatch.setattr(git_context_module, "_run_git_command", fake_run)
    monkeypatch.setattr(git_context_module, "_run_git_command_rc", fake_rc)

    off_context = get_git_context(dirty_mode="off")
    cached_context = get_git_context(dirty_mode="cached")

    assert off_context is not None
    assert "dirty" not in off_context
    assert cached_context is not None
    assert cached_context.get("dirty") is False

    show_toplevel_calls = [call for call in run_calls if call[:2] == ["rev-parse", "--show-toplevel"]]
    assert len(show_toplevel_calls) == 2
    ls_files_calls = [call for call in rc_calls if call[:1] == ["ls-files"]]
    assert len(ls_files_calls) == 1


def test_git_context_rejects_invalid_dirty_mode():
    with pytest.raises(ValueError, match="dirty_mode must be"):
        get_git_context(dirty_mode=cast(Any, "invalid"))



from __future__ import annotations

import argparse
from pathlib import Path
from types import SimpleNamespace

import pytest

from evals import run_eval
from llmdebug import _json as json


def test_execution_config_defaults_to_testcontainers() -> None:
    parser = argparse.ArgumentParser()
    run_eval.add_execution_arguments(parser)
    args = parser.parse_args([])
    config = run_eval.build_execution_config_from_args(args)
    assert config.executor == "testcontainers"
    assert config.container_image


def test_execution_config_rejects_invalid_container_env() -> None:
    parser = argparse.ArgumentParser()
    run_eval.add_execution_arguments(parser)
    args = parser.parse_args(["--container-env", "INVALID"])
    with pytest.raises(ValueError, match="expected KEY=VALUE"):
        run_eval.build_execution_config_from_args(args)


def test_run_pytest_dispatches_local_executor(tmp_path, monkeypatch) -> None:
    marker = run_eval.RunResult(returncode=0, duration_sec=0.0, stdout="ok", stderr="")
    monkeypatch.setattr(run_eval, "_run_pytest_local", lambda **kwargs: marker)
    result = run_eval.run_pytest(
        cwd=tmp_path,
        python_args=["-m", "pytest", "-q"],
        condition=run_eval.Condition.TRACEBACK_ONLY,
        timeout_sec=1,
        output_format="json_compact",
        execution=run_eval.ExecutionConfig(executor="local"),
    )
    assert result is marker


def test_run_pytest_dispatches_testcontainers_executor(tmp_path, monkeypatch) -> None:
    marker = run_eval.RunResult(returncode=1, duration_sec=0.0, stdout="", stderr="x")
    monkeypatch.setattr(run_eval, "_run_pytest_testcontainers", lambda **kwargs: marker)
    result = run_eval.run_pytest(
        cwd=tmp_path,
        python_args=["-m", "pytest", "-q"],
        condition=run_eval.Condition.TRACEBACK_ONLY,
        timeout_sec=1,
        output_format="json_compact",
        execution=run_eval.ExecutionConfig(executor="testcontainers"),
    )
    assert result is marker


def test_write_run_outputs_truncates_to_latest_lines(tmp_path) -> None:
    run = run_eval.RunResult(
        returncode=1,
        duration_sec=0.0,
        stdout="line-1\nline-2\nline-3\nline-4\n",
        stderr="err-1\nerr-2\n",
    )
    run_eval._write_run_outputs(tmp_path, run, prefix="pytest_before", tail_lines=2)

    stdout_text = (tmp_path / "pytest_before_stdout.txt").read_text(encoding="utf-8")
    stderr_text = (tmp_path / "pytest_before_stderr.txt").read_text(encoding="utf-8")
    combined_text = (tmp_path / "pytest_before_combined.txt").read_text(encoding="utf-8")

    assert "line-1" not in stdout_text
    assert "line-2" not in stdout_text
    assert "line-3\nline-4\n" in stdout_text
    assert "truncated 2 earlier line(s)" in stdout_text
    assert stderr_text == "err-1\nerr-2\n"
    assert "line-3\nline-4\n" in combined_text
    assert "err-1\nerr-2\n" in combined_text


def test_write_run_outputs_applies_hard_char_cap(tmp_path, monkeypatch) -> None:
    monkeypatch.setattr(run_eval, "PERSISTED_PYTEST_OUTPUT_HARD_CHAR_CAP", 64)
    huge_line = "prefix\n" + ("x" * 500)
    run = run_eval.RunResult(returncode=1, duration_sec=0.0, stdout=huge_line, stderr="")
    run_eval._write_run_outputs(tmp_path, run, prefix="pytest_before", tail_lines=0)

    stdout_text = (tmp_path / "pytest_before_stdout.txt").read_text(encoding="utf-8")
    assert "character(s)" in stdout_text
    assert stdout_text.endswith("x" * 64)
    assert len(stdout_text) < 200


def test_truncate_snapshot_for_artifact_caps_nested_strings() -> None:
    huge = "X" * 1000
    snapshot = {
        "exception": {"type": "AssertionError", "message": huge},
        "frames": [{"code": huge, "locals": {"value": huge}}],
    }

    truncated = run_eval._truncate_snapshot_for_artifact(snapshot, max_string_chars=80)

    assert truncated is not snapshot
    assert "snapshot string truncated" in truncated["exception"]["message"]
    assert len(truncated["exception"]["message"]) <= 80
    assert len(truncated["frames"][0]["code"]) <= 80
    assert len(truncated["frames"][0]["locals"]["value"]) <= 80
    assert snapshot["exception"]["message"] == huge


def test_truncate_snapshot_for_artifact_zero_disables_truncation() -> None:
    huge = "Y" * 1000
    snapshot = {"exception": {"message": huge}}
    truncated = run_eval._truncate_snapshot_for_artifact(snapshot, max_string_chars=0)
    assert truncated["exception"]["message"] == huge


def test_run_eval_preflights_testcontainers_runtime(tmp_path, monkeypatch, capsys) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "preflight_case"
    case_dir.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="preflight_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=["preflight"],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    monkeypatch.setattr(
        run_eval,
        "check_testcontainers_runtime_available",
        lambda: (False, "docker daemon is not reachable"),
    )

    exit_code = run_eval.main(
        [
            "--run-id",
            "preflight-run",
            "--patcher",
            "null",
            "--conditions",
            "traceback_only",
            "--k",
            "1",
        ]
    )
    assert exit_code == 2
    assert "Environment check failed: docker daemon is not reachable" in capsys.readouterr().out


def test_run_eval_prunes_stale_containers_before_testcontainers_run(tmp_path, monkeypatch) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "cleanup_case"
    case_dir.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="cleanup_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    monkeypatch.setattr(
        run_eval,
        "check_testcontainers_runtime_available",
        lambda: (True, "ok"),
    )
    cleanup_calls: dict[str, int] = {}

    def _cleanup(*, grace_sec: int, now_utc=None):
        cleanup_calls["grace_sec"] = grace_sec
        return 0, 0, None

    monkeypatch.setattr(run_eval, "cleanup_stale_eval_containers", _cleanup)
    monkeypatch.setattr(run_eval, "make_patcher", lambda _args: SimpleNamespace(name="stub"))

    def _run(*, case, condition, run_id, run_metadata, **_kwargs):
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(
        [
            "--run-id",
            "cleanup-run",
            "--patcher",
            "null",
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "testcontainers",
        ]
    )

    assert exit_code == 0
    assert cleanup_calls == {"grace_sec": run_eval.DEFAULT_CONTAINER_CLEANUP_STALE_GRACE_SEC}


def test_run_eval_openai_max_propose_default_in_patcher_config(tmp_path, monkeypatch) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "openai_default_budget_case"
    case_dir.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="openai_default_budget_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    monkeypatch.setattr(run_eval, "make_patcher", lambda _args: SimpleNamespace(name="openai"))

    captured_run_metadata: dict[str, object] = {}

    def _run(*, case, condition, run_id, run_metadata, **_kwargs):
        captured_run_metadata.update(dict(run_metadata))
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(
        [
            "--run-id",
            "openai-default-budget",
            "--patcher",
            "openai",
            "--openai-model",
            "dummy-model",
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "local",
        ]
    )

    assert exit_code == 0
    patcher_config = captured_run_metadata.get("patcher_config")
    assert isinstance(patcher_config, dict)
    assert patcher_config["openai_max_propose_sec"] == 180.0
    assert patcher_config["openai_http2"] is True
    assert patcher_config["openai_transport_max_retries"] == 6
    assert patcher_config["openai_transport_retry_backoff_base_sec"] == 1.0
    assert patcher_config["openai_transport_retry_backoff_max_sec"] == 8.0


def test_run_eval_openai_transport_retry_settings_in_patcher_config(tmp_path, monkeypatch) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "openai_transport_budget_case"
    case_dir.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="openai_transport_budget_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    monkeypatch.setattr(run_eval, "make_patcher", lambda _args: SimpleNamespace(name="openai"))

    captured_run_metadata: dict[str, object] = {}

    def _run(*, case, condition, run_id, run_metadata, **_kwargs):
        captured_run_metadata.update(dict(run_metadata))
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(
        [
            "--run-id",
            "openai-transport-settings",
            "--patcher",
            "openai",
            "--openai-model",
            "dummy-model",
            "--openai-transport-max-retries",
            "9",
            "--openai-transport-retry-backoff-base-sec",
            "1.5",
            "--openai-transport-retry-backoff-max-sec",
            "12.0",
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "local",
        ]
    )

    assert exit_code == 0
    patcher_config = captured_run_metadata.get("patcher_config")
    assert isinstance(patcher_config, dict)
    assert patcher_config["openai_transport_max_retries"] == 9
    assert patcher_config["openai_http2"] is True
    assert patcher_config["openai_transport_retry_backoff_base_sec"] == 1.5
    assert patcher_config["openai_transport_retry_backoff_max_sec"] == 12.0


def test_run_eval_openai_http2_disable_in_patcher_config(tmp_path, monkeypatch) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "openai_http2_case"
    case_dir.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="openai_http2_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    monkeypatch.setattr(run_eval, "make_patcher", lambda _args: SimpleNamespace(name="openai"))

    captured_run_metadata: dict[str, object] = {}

    def _run(*, case, condition, run_id, run_metadata, **_kwargs):
        captured_run_metadata.update(dict(run_metadata))
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(
        [
            "--run-id",
            "openai-http2-off",
            "--patcher",
            "openai",
            "--openai-model",
            "dummy-model",
            "--no-openai-http2",
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "local",
        ]
    )

    assert exit_code == 0
    patcher_config = captured_run_metadata.get("patcher_config")
    assert isinstance(patcher_config, dict)
    assert patcher_config["openai_http2"] is False


def test_run_eval_case_list_file_filters_selected_cases(tmp_path, monkeypatch) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_root = repo_root / "synthetic_cases"
    case_root.mkdir(parents=True, exist_ok=True)
    (case_root / "case_a").mkdir(parents=True, exist_ok=True)
    (case_root / "case_b").mkdir(parents=True, exist_ok=True)
    (case_root / "case_c").mkdir(parents=True, exist_ok=True)

    case_a = run_eval.CaseSpec(
        case_id="case_a",
        path=case_root / "case_a",
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )
    case_b = run_eval.CaseSpec(
        case_id="case_b",
        path=case_root / "case_b",
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )
    case_c = run_eval.CaseSpec(
        case_id="case_c",
        path=case_root / "case_c",
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    case_list_file = tmp_path / "selected_cases.txt"
    case_list_file.write_text("# comment\n\ncase_b\n", encoding="utf-8")

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case_a, case_b, case_c])
    monkeypatch.setattr(run_eval, "make_patcher", lambda _args: SimpleNamespace(name="stub"))

    executed_case_ids: list[str] = []

    def _run(*, case, condition, run_id, run_metadata, **_kwargs):
        executed_case_ids.append(case.case_id)
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(
        [
            "--run-id",
            "case-list-filter",
            "--patcher",
            "null",
            "--case",
            "case_a",
            "--case-list-file",
            str(case_list_file),
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "local",
        ]
    )

    assert exit_code == 0
    assert sorted(executed_case_ids) == ["case_a", "case_b"]


def test_run_eval_case_list_file_rejects_unknown_case_id(tmp_path, monkeypatch, capsys) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_root = repo_root / "synthetic_cases"
    case_root.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="known_case",
        path=case_root / "known_case",
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    case_list_file = tmp_path / "selected_cases.txt"
    case_list_file.write_text("unknown_case\n", encoding="utf-8")

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])

    exit_code = run_eval.main(
        [
            "--run-id",
            "case-list-unknown",
            "--patcher",
            "null",
            "--case-list-file",
            str(case_list_file),
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "local",
        ]
    )

    assert exit_code == 2
    assert "Unknown case id(s) requested via --case/--case-list-file" in capsys.readouterr().out


def test_run_eval_case_list_file_read_error_returns_cli_error(
    tmp_path, monkeypatch, capsys
) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_root = repo_root / "synthetic_cases"
    case_root.mkdir(parents=True, exist_ok=True)
    case_dir = case_root / "known_case"
    case_dir.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="known_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    case_list_file = tmp_path / "selected_cases.txt"
    case_list_file.write_text("known_case\n", encoding="utf-8")

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])

    original_read_text = Path.read_text

    def _fake_read_text(path: Path, *args, **kwargs) -> str:
        if path == case_list_file:
            raise OSError("permission denied")
        return original_read_text(path, *args, **kwargs)

    monkeypatch.setattr(Path, "read_text", _fake_read_text)

    exit_code = run_eval.main(
        [
            "--run-id",
            "case-list-read-error",
            "--patcher",
            "null",
            "--case-list-file",
            str(case_list_file),
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "local",
        ]
    )

    assert exit_code == 2
    assert "Failed to read case-list file" in capsys.readouterr().out


def test_run_eval_forwards_pytest_output_tail_lines(tmp_path, monkeypatch) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "tail_lines_forward_case"
    case_dir.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="tail_lines_forward_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    monkeypatch.setattr(run_eval, "make_patcher", lambda _args: SimpleNamespace(name="stub"))

    observed_tail_lines: dict[str, int] = {}

    def _run(*, case, condition, run_id, run_metadata, pytest_output_tail_lines, **_kwargs):
        observed_tail_lines["value"] = int(pytest_output_tail_lines)
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(
        [
            "--run-id",
            "tail-lines-forward",
            "--patcher",
            "null",
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "local",
            "--pytest-output-tail-lines",
            "123",
        ]
    )

    assert exit_code == 0
    assert observed_tail_lines["value"] == 123


def test_run_eval_forwards_snapshot_artifact_max_string_chars(tmp_path, monkeypatch) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "snapshot_cap_forward_case"
    case_dir.mkdir(parents=True, exist_ok=True)

    case = run_eval.CaseSpec(
        case_id="snapshot_cap_forward_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    monkeypatch.setattr(run_eval, "make_patcher", lambda _args: SimpleNamespace(name="stub"))

    observed_snapshot_cap: dict[str, int] = {}

    def _run(
        *,
        case,
        condition,
        run_id,
        run_metadata,
        snapshot_artifact_max_string_chars,
        **_kwargs,
    ):
        observed_snapshot_cap["value"] = int(snapshot_artifact_max_string_chars)
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(
        [
            "--run-id",
            "snapshot-cap-forward",
            "--patcher",
            "null",
            "--conditions",
            "traceback_only",
            "--k",
            "1",
            "--executor",
            "local",
            "--snapshot-artifact-max-string-chars",
            "321",
        ]
    )

    assert exit_code == 0
    assert observed_snapshot_cap["value"] == 321


def test_run_eval_config_precedence_and_resolved_lockfile(tmp_path, monkeypatch) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "config_precedence_case"
    case_dir.mkdir(parents=True, exist_ok=True)
    config_path = tmp_path / "run_eval.toml"
    config_path.write_text(
        """
[run_eval]
patcher = "null"
conditions = "traceback_only"
k = 3
run_id = "run-from-config"
executor = "local"
""".strip()
        + "\n",
        encoding="utf-8",
    )

    case = run_eval.CaseSpec(
        case_id="config_precedence_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    monkeypatch.setattr(run_eval, "make_patcher", lambda _args: SimpleNamespace(name="stub"))
    observed: dict[str, object] = {}

    def _run(*, case, condition, run_id, run_metadata, k, **_kwargs):
        observed["run_id"] = run_id
        observed["k"] = int(k)
        observed["run_metadata"] = dict(run_metadata)
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(
        [
            "--config",
            str(config_path),
            "--run-id",
            "run-from-cli",
            "--k",
            "1",
            "--executor",
            "local",
        ]
    )

    assert exit_code == 0
    assert observed["run_id"] == "run-from-cli"
    assert observed["k"] == 1
    lock_path = repo_root / "evals" / "artifacts" / "configs" / "run-from-cli.resolved.json"
    assert lock_path.exists()
    lock_payload = json.loads(lock_path.read_text(encoding="utf-8"))
    assert lock_payload["run_id"] == "run-from-cli"
    assert lock_payload["config_files"] == [str(config_path.resolve())]
    assert lock_payload["resolved_args"]["k"] == 1
    assert lock_payload["value_sources"]["k"] == "cli"
    assert str(lock_payload["value_sources"]["patcher"]).startswith("config:")


def test_run_eval_rejects_plaintext_openai_api_key_in_config(tmp_path, capsys) -> None:
    config_path = tmp_path / "run_eval.toml"
    config_path.write_text(
        """
[run_eval]
patcher = "openai"
openai_model = "demo"
openai_api_key = "sk-test-literal"
""".strip()
        + "\n",
        encoding="utf-8",
    )

    exit_code = run_eval.main(["--config", str(config_path)])
    assert exit_code == 2
    assert "openai_api_key" in capsys.readouterr().out


def test_run_eval_resolves_openai_api_key_env_from_config_and_redacts_lock(
    tmp_path, monkeypatch
) -> None:
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)
    case_dir = repo_root / "synthetic_cases" / "config_env_secret_case"
    case_dir.mkdir(parents=True, exist_ok=True)
    config_path = tmp_path / "run_eval.toml"
    config_path.write_text(
        """
[run_eval]
run_id = "config-env-secret-run"
patcher = "openai"
conditions = "traceback_only"
executor = "local"
k = 1

[run_eval.openai]
model = "demo"
api_key_env = "LLMDEBUG_TEST_OPENAI_KEY"
""".strip()
        + "\n",
        encoding="utf-8",
    )
    monkeypatch.setenv("LLMDEBUG_TEST_OPENAI_KEY", "super-secret-token")

    case = run_eval.CaseSpec(
        case_id="config_env_secret_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )

    monkeypatch.setattr(run_eval, "_REPO_ROOT", repo_root)
    monkeypatch.setattr(run_eval, "load_cases", lambda _cases_dir: [case])
    observed_openai_key: dict[str, str] = {}

    def _make_patcher(_args):
        observed_openai_key["value"] = str(_args.openai_api_key)
        return SimpleNamespace(name="openai")

    monkeypatch.setattr(run_eval, "make_patcher", _make_patcher)

    def _run(*, case, condition, run_id, run_metadata, **_kwargs):
        return [
            {
                "run_id": run_id,
                "case_id": case.case_id,
                "condition": condition.value,
                "attempt": 1,
                "success": False,
                "run_metadata": dict(run_metadata),
            }
        ]

    monkeypatch.setattr(run_eval, "run_one_condition", _run)

    exit_code = run_eval.main(["--config", str(config_path)])
    assert exit_code == 0
    assert observed_openai_key["value"] == "super-secret-token"

    lock_path = (
        repo_root / "evals" / "artifacts" / "configs" / "config-env-secret-run.resolved.json"
    )
    lock_payload = json.loads(lock_path.read_text(encoding="utf-8"))
    assert lock_payload["resolved_args"]["openai_api_key"] == run_eval.RUN_EVAL_LOCKFILE_REDACTED



from __future__ import annotations

import argparse
from pathlib import Path
from typing import Any

from . import _json as json


def _load_jsonl(path: Path) -> list[dict[str, Any]]:
    rows: list[dict[str, Any]] = []
    for line_no, line in enumerate(path.read_text(encoding="utf-8").splitlines(), start=1):
        stripped = line.strip()
        if not stripped:
            continue
        try:
            payload = json.loads(stripped)
        except json.JSONDecodeError as exc:
            raise ValueError(f"{path}:{line_no}: invalid JSON: {exc.msg}") from exc
        if not isinstance(payload, dict):
            raise ValueError(f"{path}:{line_no}: expected JSON object row")
        rows.append(payload)
    return rows


def _coerce_attempt(value: Any) -> int:
    try:
        return max(0, int(value))
    except (TypeError, ValueError):
        return 0


def _terminal_by_case(
    *,
    rows: list[dict[str, Any]],
    condition: str,
) -> dict[str, dict[str, Any]]:
    terminal: dict[str, dict[str, Any]] = {}
    for row in rows:
        if str(row.get("condition") or "") != condition:
            continue
        case_id = str(row.get("case_id") or "").strip()
        if not case_id:
            continue
        attempt = _coerce_attempt(row.get("attempt"))
        previous = terminal.get(case_id)
        if previous is None:
            terminal[case_id] = row
            continue
        prev_attempt = _coerce_attempt(previous.get("attempt"))
        if attempt > prev_attempt or attempt == prev_attempt:
            # Last-write-wins for duplicate rows at the same attempt number.
            terminal[case_id] = row
    return terminal


def _select_case_ids(
    *,
    terminal_rows: dict[str, dict[str, Any]],
    mode: str,
    attempt_threshold: int,
) -> list[str]:
    selected: list[str] = []
    threshold = max(1, int(attempt_threshold))
    for case_id, row in terminal_rows.items():
        success = bool(row.get("success"))
        attempt = _coerce_attempt(row.get("attempt"))
        is_failed = not success
        is_attempt_ge = attempt >= threshold
        include = False
        if mode == "failed":
            include = is_failed
        elif mode == "attempt-ge":
            include = is_attempt_ge
        elif mode == "failed-or-attempt-ge":
            include = is_failed or is_attempt_ge
        if include:
            selected.append(case_id)
    return sorted(selected)


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(
        description="Select rerun case IDs from eval JSONL outputs."
    )
    parser.add_argument(
        "--results-jsonl",
        required=True,
        help="Path to eval JSONL file (attempt-level records).",
    )
    parser.add_argument(
        "--condition",
        default="traceback_only",
        help="Condition to read terminal rows from (default: traceback_only).",
    )
    parser.add_argument(
        "--mode",
        choices=["failed", "attempt-ge", "failed-or-attempt-ge"],
        default="failed",
        help="Selection mode (default: failed).",
    )
    parser.add_argument(
        "--attempt-threshold",
        type=int,
        default=2,
        help="Attempt threshold for attempt-ge modes (default: 2).",
    )
    parser.add_argument(
        "--out",
        required=True,
        help="Output path for newline-delimited case IDs.",
    )
    args = parser.parse_args(argv)

    input_path = Path(str(args.results_jsonl)).expanduser()
    if not input_path.exists() or not input_path.is_file():
        print(f"Results JSONL not found or not a file: {input_path}")
        return 2

    output_path = Path(str(args.out)).expanduser()
    try:
        rows = _load_jsonl(input_path)
    except ValueError as exc:
        print(str(exc))
        return 2

    condition = str(args.condition).strip()
    if not condition:
        print("Invalid --condition: must be a non-empty string")
        return 2

    terminal_rows = _terminal_by_case(rows=rows, condition=condition)
    if not terminal_rows:
        print(f"No rows found for condition={condition!r} in {input_path}")
        return 2

    selected_case_ids = _select_case_ids(
        terminal_rows=terminal_rows,
        mode=str(args.mode),
        attempt_threshold=int(args.attempt_threshold),
    )
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(
        "".join(f"{case_id}\n" for case_id in selected_case_ids),
        encoding="utf-8",
    )

    success_count = sum(1 for row in terminal_rows.values() if bool(row.get("success")))
    print(
        "Selected "
        f"{len(selected_case_ids)} case(s) from {len(terminal_rows)} terminal row(s) "
        f"for condition={condition!r}, mode={args.mode!r}, "
        f"attempt_threshold={int(args.attempt_threshold)} (terminal successes={success_count})."
    )
    print(f"Wrote case list: {output_path}")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Additional edge-case coverage for output helpers."""

from __future__ import annotations

import contextlib
import os
import re
import time
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import pytest

from llmdebug import _json as json
from llmdebug.config import SnapshotConfig
from llmdebug.output import get_latest_snapshot, write_bundle


def test_timestamp_format() -> None:
    from llmdebug.output import _timestamp

    stamp = _timestamp()
    assert re.fullmatch(r"\d{8}T\d{6}Z", stamp)


def test_directory_lock_timeout_branch(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    from llmdebug import output

    class _FakeLock:
        def __init__(self, _path: Path, timeout: float) -> None:
            self.timeout = timeout
            self.is_locked = False

        def acquire(self) -> None:
            raise output.Timeout(str(self.timeout))

        def release(self) -> None:
            self.is_locked = False

    monkeypatch.setattr(output, "FileLock", _FakeLock)

    with output._directory_lock(tmp_path, timeout=0.01) as acquired:
        assert acquired is False


def test_write_bundle_warns_when_lock_not_acquired(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    from llmdebug import output

    @contextlib.contextmanager
    def _never_lock(_out_dir: Path, _timeout: float) -> Iterator[bool]:
        yield False

    monkeypatch.setattr(output, "_directory_lock", _never_lock)
    cfg = SnapshotConfig(out_dir=str(tmp_path), output_format="json")
    path = write_bundle({"name": "lock-timeout", "frames": []}, cfg)

    assert path.exists()
    assert "lock timeout" in capsys.readouterr().err


@pytest.mark.skipif(os.name == "nt", reason="POSIX-only symlink behavior")
def test_update_latest_symlink_posix_success(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    from llmdebug import output

    target = tmp_path / "snapshot.json"
    target.write_text("{}", encoding="utf-8")

    monkeypatch.setattr(output.sys, "platform", "darwin")
    output._update_latest_symlink(tmp_path, target.name, ".json")

    latest = tmp_path / "latest.json"
    assert latest.exists()
    assert latest.is_symlink()
    assert latest.resolve() == target.resolve()


@pytest.mark.skipif(os.name == "nt", reason="POSIX-only symlink behavior")
def test_update_latest_symlink_posix_failure_cleans_temp(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    from llmdebug import output

    target = tmp_path / "snapshot.json"
    target.write_text("{}", encoding="utf-8")

    monkeypatch.setattr(output.sys, "platform", "darwin")
    real_rename = output.Path.rename

    def _failing_rename(self: Path, dst: Path) -> Path:
        if self.name.startswith(".latest.json.tmp."):
            raise OSError("rename failed")
        return real_rename(self, dst)

    monkeypatch.setattr(output.Path, "rename", _failing_rename)
    output._update_latest_symlink(tmp_path, target.name, ".json")

    assert "could not update latest.json" in capsys.readouterr().err
    assert list(tmp_path.glob(".latest.json.tmp.*")) == []


def test_snapshot_index_entry_handles_missing_path_stat() -> None:
    from llmdebug.output import _snapshot_index_entry

    missing = Path("/tmp/llmdebug-this-file-does-not-exist.json")
    payload = {
        "name": "entry-test",
        "exception": {"type": "ValueError", "message": "boom"},
        "frames": [{"file": "app.py", "line": 12}],
    }
    entry = _snapshot_index_entry(missing, payload)
    assert entry["mtime"] == 0.0
    assert entry["name"] == "entry-test"
    assert entry["exc_type"] == "ValueError"
    assert entry["file"] == "app.py"


def test_get_latest_snapshot_falls_back_from_invalid_toon_to_json(
    tmp_path: Path,
) -> None:
    latest_json = tmp_path / "latest.json"
    latest_json.write_text(
        json.dumps({"name": "json-fallback", "exception": {"type": "E"}, "frames": []}),
        encoding="utf-8",
    )
    latest_toon = tmp_path / "latest.toon"
    latest_toon.write_text("not-toon", encoding="utf-8")
    now = time.time()
    os.utime(latest_json, (now - 10, now - 10))
    os.utime(latest_toon, (now + 10, now + 10))

    snapshot = get_latest_snapshot(str(tmp_path))
    assert snapshot is not None
    assert snapshot["name"] == "json-fallback"


def test_get_latest_snapshot_warns_when_latest_json_is_not_dict(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    (tmp_path / "latest.json").write_text(json.dumps(["bad", "payload"]), encoding="utf-8")

    assert get_latest_snapshot(str(tmp_path)) is None
    assert "contains list, expected dict" in capsys.readouterr().err


def test_get_latest_snapshot_warns_on_compact_expansion_failure(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    import llmdebug.compact_keys as compact_keys

    (tmp_path / "latest.json").write_text(
        json.dumps({"__llmdebug_compact__": 1, "_exc": {"_t": "ValueError"}}),
        encoding="utf-8",
    )

    monkeypatch.setattr(compact_keys, "expand_keys", lambda _data: (_ for _ in ()).throw(ValueError("bad")))

    assert get_latest_snapshot(str(tmp_path)) is None
    assert "compact key expansion failed" in capsys.readouterr().err


def test_get_latest_snapshot_warns_on_corrupted_json(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    (tmp_path / "latest.json").write_text("{not-json", encoding="utf-8")

    assert get_latest_snapshot(str(tmp_path)) is None
    assert "is corrupted" in capsys.readouterr().err



"""Pure RCA domain engine.

This module computes RCA attempt records without performing persistence I/O.
"""

from __future__ import annotations

from typing import Any

from .gates import evaluate_gate, normalize_gate_mode
from .rca_transitions import (
    allowed_actions_for_state,
    apply_stall_rule,
    build_transition_history,
    default_next_required,
    derive_attempt_progress,
    normalize_action,
    resolve_attempt_values,
    resolve_merged_evidence_refs,
    resolve_previous_values,
    resolve_target_state,
)


def build_rca_attempt_record(
    *,
    previous: dict[str, Any] | None,
    attempt_id: str,
    session_id: str,
    action: str,
    failure_signature: str,
    evidence_refs: list[str] | None = None,
    hypothesis: str | None = None,
    experiment_plan: str | None = None,
    proposed_change_summary: str | None = None,
    verification_result: dict[str, Any] | None = None,
    gate_mode: str | None = None,
    exploratory: bool = False,
    override_reason: str | None = None,
    stall_threshold: int = 3,
    exploration_budget: int = 2,
    now: str,
) -> dict[str, Any]:
    normalized_mode = normalize_gate_mode(gate_mode)
    normalized_action = normalize_action(action)
    (
        previous_state,
        previous_hypothesis,
        previous_plan,
        previous_refs,
        previous_verification,
        previous_change_summary,
    ) = resolve_previous_values(previous)

    merged_refs = resolve_merged_evidence_refs(previous_refs, evidence_refs)
    (
        hypothesis_value,
        plan_value,
        verification_value,
        proposed_change_value,
    ) = resolve_attempt_values(
        previous_hypothesis=previous_hypothesis,
        previous_plan=previous_plan,
        previous_change_summary=previous_change_summary,
        previous_verification=previous_verification,
        hypothesis=hypothesis,
        experiment_plan=experiment_plan,
        proposed_change_summary=proposed_change_summary,
        verification_result=verification_result,
    )
    same_signature, repeat_count, hypothesis_delta, plan_advanced, hypothesis_advanced = (
        derive_attempt_progress(
            previous=previous,
            failure_signature=failure_signature,
            previous_hypothesis=previous_hypothesis,
            hypothesis_value=hypothesis_value,
            previous_plan=previous_plan,
            plan_value=plan_value,
        )
    )

    gate_payload = {
        "evidence_refs": merged_refs,
        "hypothesis": hypothesis_value,
        "experiment_plan": plan_value,
        "verification_result": verification_value,
        "exploratory": exploratory,
        "override_reason": override_reason or "",
        "hypothesis_advanced": hypothesis_advanced,
    }
    gate = evaluate_gate(
        mode=normalized_mode,
        current_state=previous_state,
        action=normalized_action,
        payload=gate_payload,
        repeated_signature_count=repeat_count,
        exploration_budget=exploration_budget,
    )

    target_state = resolve_target_state(previous_state, normalized_action, gate.allowed)
    target_state, gate_warnings = apply_stall_rule(
        target_state=target_state,
        gate_warnings=list(gate.warnings),
        same_signature=same_signature,
        repeat_count=repeat_count,
        stall_threshold=stall_threshold,
        hypothesis_advanced=hypothesis_advanced,
        plan_advanced=plan_advanced,
        normalized_action=normalized_action,
    )
    transition_history = build_transition_history(
        previous=previous,
        now=now,
        normalized_action=normalized_action,
        previous_state=previous_state,
        target_state=target_state,
        gate_allowed=gate.allowed,
        gate_mode=gate.mode,
    )
    created_at = str(previous.get("created_at") or now) if previous else now
    next_required = (
        list(gate.next_required) if gate.next_required else default_next_required(target_state)
    )

    return {
        "attempt_id": attempt_id,
        "session_id": session_id,
        "state": target_state,
        "failure_signature": failure_signature,
        "repeat_count": repeat_count,
        "evidence_refs": merged_refs,
        "hypothesis": hypothesis_value or None,
        "hypothesis_delta": hypothesis_delta,
        "experiment_plan": plan_value or None,
        "proposed_change_summary": proposed_change_value or None,
        "verification_result": verification_value,
        "gate_feedback": gate.to_dict(),
        "gate_warnings": gate_warnings,
        "allowed_actions": allowed_actions_for_state(target_state),
        "next_required": next_required,
        "exploratory": exploratory,
        "transition_history": transition_history,
        "created_at": created_at,
        "updated_at": now,
    }



from __future__ import annotations

import csv

from evals.export_policy_dataset import main as export_policy_dataset_main
from llmdebug import _json as json


def test_export_policy_dataset_writes_flat_csv_and_dedupes(tmp_path) -> None:
    events_path = tmp_path / "events.jsonl"
    output_csv = tmp_path / "dataset.csv"
    events = [
        {
            "schema_version": 1,
            "run_id": "run_a",
            "case_id": "case_1",
            "condition": "adaptive_llm_discretion",
            "attempt": 1,
            "note": "no_patch_proposed",
            "decision_id": "d1",
            "decision_index": 1,
            "mode": "on",
            "algorithm": "linucb",
            "selected_action": "invoke_helper",
            "applied_action": "invoke_helper",
            "default_action": "invoke_helper",
            "propensity": 0.8,
            "decision_strategy": "linucb_greedy",
            "scores": {"skip_helper": 0.2, "invoke_helper": 0.5},
            "feature_values": {"bias": 1.0, "attempt_frac": 0.5},
            "reward": 0.2,
            "reward_success": 0.0,
            "reward_token_penalty": 0.5,
            "reward_latency_penalty": 0.3,
            "success": False,
            "tokens_total": 321,
            "attempt_wall_sec": 4.2,
            "override_reason": "",
            "model_updated": True,
            "update_reason": "on_policy_update",
            "created_at_utc": "2026-02-26T00:00:00Z",
        },
        {
            "schema_version": 1,
            "run_id": "run_a",
            "case_id": "case_1",
            "condition": "adaptive_llm_discretion",
            "attempt": 1,
            "note": "no_patch_proposed",
            "decision_id": "d1",
            "decision_index": 1,
            "mode": "on",
            "algorithm": "linucb",
            "selected_action": "invoke_helper",
            "applied_action": "invoke_helper",
            "default_action": "invoke_helper",
            "propensity": 0.8,
            "decision_strategy": "linucb_greedy",
            "scores": {"skip_helper": 0.2, "invoke_helper": 0.5},
            "feature_values": {"bias": 1.0, "attempt_frac": 0.5},
            "reward": 0.9,
            "reward_success": 1.0,
            "reward_token_penalty": 0.05,
            "reward_latency_penalty": 0.05,
            "success": True,
            "tokens_total": 100,
            "attempt_wall_sec": 1.0,
            "override_reason": "",
            "model_updated": True,
            "update_reason": "on_policy_update",
            "created_at_utc": "2026-02-26T00:00:01Z",
        },
        {
            "schema_version": 1,
            "run_id": "run_a",
            "case_id": "case_2",
            "condition": "adaptive_llm_discretion",
            "attempt": 1,
            "note": "tests_still_failing",
            "decision_id": "d2",
            "decision_index": 2,
            "mode": "shadow",
            "algorithm": "linucb",
            "selected_action": "skip_helper",
            "applied_action": "invoke_helper",
            "default_action": "invoke_helper",
            "propensity": 0.0,
            "decision_strategy": "epsilon_uniform",
            "scores": {"skip_helper": 0.4, "invoke_helper": 0.3},
            "feature_values": {"bias": 1.0, "attempt_frac": 1.0},
            "reward": -0.2,
            "reward_success": 0.0,
            "reward_token_penalty": 0.1,
            "reward_latency_penalty": 0.1,
            "success": False,
            "tokens_total": 50,
            "attempt_wall_sec": 2.0,
            "override_reason": "shadow_mode_baseline",
            "model_updated": False,
            "update_reason": "shadow_no_update",
            "created_at_utc": "2026-02-26T00:00:02Z",
        },
    ]
    events_path.write_text(
        "\n".join(json.dumps(event, ensure_ascii=False) for event in events) + "\n",
        encoding="utf-8",
    )

    exit_code = export_policy_dataset_main(
        [
            "--events",
            str(events_path),
            "--output-csv",
            str(output_csv),
            "--min-propensity",
            "0.1",
        ]
    )

    assert exit_code == 0
    with output_csv.open("r", encoding="utf-8", newline="") as handle:
        rows = list(csv.DictReader(handle))

    assert len(rows) == 2
    by_decision = {row["decision_id"]: row for row in rows}
    assert by_decision["d1"]["reward"] == "0.9"
    assert by_decision["d1"]["feature_bias"] == "1.0"
    assert by_decision["d1"]["feature_attempt_frac"] == "0.5"
    assert by_decision["d1"]["importance_weight"] == "1.25"
    assert by_decision["d2"]["propensity_clipped"] == "0.1"
    assert by_decision["d2"]["importance_weight"] == "10.0"


def test_export_policy_dataset_strict_input_fails_fast(tmp_path) -> None:
    events_path = tmp_path / "events_invalid.jsonl"
    events_path.write_text('{"run_id":"r1"}\n{invalid-json\n', encoding="utf-8")

    exit_code = export_policy_dataset_main(
        [
            "--events",
            str(events_path),
            "--strict-input",
        ]
    )

    assert exit_code == 2



from __future__ import annotations

import random
import threading
import uuid
from collections.abc import Mapping
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

import numpy as np

from . import _json as json

ACTION_SKIP_HELPER = "skip_helper"
ACTION_INVOKE_HELPER = "invoke_helper"
ONLINE_POLICY_ACTIONS = (ACTION_SKIP_HELPER, ACTION_INVOKE_HELPER)
ONLINE_POLICY_MODES = {"off", "shadow", "on"}
ONLINE_POLICY_ALGOS = {"linucb"}
STATE_SCHEMA_VERSION = 1


def _utc_now() -> str:
    return (
        datetime.now(timezone.utc).replace(microsecond=0).isoformat().replace("+00:00", "Z")
    )


def _coerce_float(value: Any, *, default: float) -> float:
    try:
        return float(value)
    except (TypeError, ValueError):
        return float(default)


def _coerce_nonnegative_int(value: Any, *, default: int) -> int:
    try:
        out = int(value)
    except (TypeError, ValueError):
        return int(default)
    return max(0, out)


def _safe_float(value: Any, *, default: float = 0.0) -> float:
    try:
        out = float(value)
    except (TypeError, ValueError):
        return float(default)
    if np.isnan(out) or np.isinf(out):
        return float(default)
    return out


@dataclass(frozen=True)
class OnlinePolicyConfig:
    mode: str = "off"
    algorithm: str = "linucb"
    alpha: float = 0.8
    epsilon: float = 0.05
    l2: float = 1.0
    warmup_decisions: int = 200
    save_every: int = 50
    state_path: Path | None = None
    events_path: Path | None = None
    random_seed: int | None = None

    @classmethod
    def normalize(
        cls,
        *,
        mode: str = "off",
        algorithm: str = "linucb",
        alpha: float = 0.8,
        epsilon: float = 0.05,
        l2: float = 1.0,
        warmup_decisions: int = 200,
        save_every: int = 50,
        state_path: Path | None = None,
        events_path: Path | None = None,
        random_seed: int | None = None,
    ) -> OnlinePolicyConfig:
        mode_norm = str(mode).strip().lower()
        if mode_norm not in ONLINE_POLICY_MODES:
            mode_norm = "off"
        algo_norm = str(algorithm).strip().lower()
        if algo_norm not in ONLINE_POLICY_ALGOS:
            algo_norm = "linucb"
        alpha_norm = max(0.0, _coerce_float(alpha, default=0.8))
        epsilon_norm = min(1.0, max(0.0, _coerce_float(epsilon, default=0.05)))
        l2_norm = max(1e-9, _coerce_float(l2, default=1.0))
        warmup_norm = _coerce_nonnegative_int(warmup_decisions, default=200)
        save_every_norm = max(1, _coerce_nonnegative_int(save_every, default=50))
        seed_norm: int | None = None
        if random_seed is not None:
            try:
                seed_norm = int(random_seed)
            except (TypeError, ValueError):
                seed_norm = None
        return cls(
            mode=mode_norm,
            algorithm=algo_norm,
            alpha=alpha_norm,
            epsilon=epsilon_norm,
            l2=l2_norm,
            warmup_decisions=warmup_norm,
            save_every=save_every_norm,
            state_path=Path(state_path) if state_path is not None else None,
            events_path=Path(events_path) if events_path is not None else None,
            random_seed=seed_norm,
        )


@dataclass(frozen=True)
class PolicyDecision:
    decision_id: str
    mode: str
    algorithm: str
    decision_index: int
    created_at_utc: str
    selected_action: str
    applied_action: str
    default_action: str
    propensity: float
    decision_strategy: str
    scores: dict[str, float]
    feature_values: dict[str, float]


@dataclass(frozen=True)
class PolicyEvent:
    run_id: str
    case_id: str
    condition: str
    attempt: int
    note: str
    decision_id: str
    mode: str
    algorithm: str
    decision_index: int
    selected_action: str
    applied_action: str
    default_action: str
    propensity: float
    decision_strategy: str
    scores: dict[str, float]
    feature_values: dict[str, float]
    reward: float
    reward_success: float
    reward_token_penalty: float
    reward_latency_penalty: float
    success: bool
    tokens_total: int
    attempt_wall_sec: float
    override_reason: str
    model_updated: bool
    update_reason: str
    created_at_utc: str

    def to_json(self) -> dict[str, Any]:
        return {
            "schema_version": STATE_SCHEMA_VERSION,
            "run_id": self.run_id,
            "case_id": self.case_id,
            "condition": self.condition,
            "attempt": self.attempt,
            "note": self.note,
            "decision_id": self.decision_id,
            "mode": self.mode,
            "algorithm": self.algorithm,
            "decision_index": self.decision_index,
            "selected_action": self.selected_action,
            "applied_action": self.applied_action,
            "default_action": self.default_action,
            "propensity": self.propensity,
            "decision_strategy": self.decision_strategy,
            "scores": dict(self.scores),
            "feature_values": dict(self.feature_values),
            "reward": self.reward,
            "reward_success": self.reward_success,
            "reward_token_penalty": self.reward_token_penalty,
            "reward_latency_penalty": self.reward_latency_penalty,
            "success": self.success,
            "tokens_total": self.tokens_total,
            "attempt_wall_sec": self.attempt_wall_sec,
            "override_reason": self.override_reason,
            "model_updated": self.model_updated,
            "update_reason": self.update_reason,
            "created_at_utc": self.created_at_utc,
        }


@dataclass
class LinUCBState:
    feature_names: list[str] = field(default_factory=list)
    decisions: int = 0
    updates: int = 0
    action_counts: dict[str, int] = field(
        default_factory=lambda: dict.fromkeys(ONLINE_POLICY_ACTIONS, 0)
    )
    A: dict[str, np.ndarray] = field(default_factory=dict)
    b: dict[str, np.ndarray] = field(default_factory=dict)


class AdaptiveOnlinePolicyController:
    def __init__(self, config: OnlinePolicyConfig) -> None:
        self.config = config
        self._lock = threading.RLock()
        self._rng = random.Random(config.random_seed)
        self._state = LinUCBState()
        self._save_counter = 0
        self.load_error: str | None = None
        if self.config.algorithm == "linucb":
            self._load_state_if_present()

    @property
    def mode(self) -> str:
        return self.config.mode

    @property
    def algorithm(self) -> str:
        return self.config.algorithm

    @property
    def state(self) -> LinUCBState:
        with self._lock:
            return LinUCBState(
                feature_names=list(self._state.feature_names),
                decisions=int(self._state.decisions),
                updates=int(self._state.updates),
                action_counts={k: int(v) for k, v in self._state.action_counts.items()},
                A={k: v.copy() for k, v in self._state.A.items()},
                b={k: v.copy() for k, v in self._state.b.items()},
            )

    def decide(
        self,
        *,
        feature_values: Mapping[str, float],
        default_action: str = ACTION_INVOKE_HELPER,
    ) -> PolicyDecision:
        cleaned_features = {
            str(k): _safe_float(v, default=0.0) for k, v in dict(feature_values).items()
        }
        with self._lock:
            self._ensure_feature_state(cleaned_features=cleaned_features)
            decision_index = self._state.decisions + 1
            mode = self.config.mode
            default_norm = (
                default_action if default_action in ONLINE_POLICY_ACTIONS else ACTION_INVOKE_HELPER
            )
            if mode == "off":
                selected = default_norm
                applied = default_norm
                propensity = 1.0
                strategy = "policy_off"
                scores = {
                    ACTION_SKIP_HELPER: 0.0,
                    ACTION_INVOKE_HELPER: 0.0,
                }
            else:
                x = self._feature_vector(cleaned_features=cleaned_features)
                scores = self._linucb_scores(x=x)
                greedy = max(
                    ONLINE_POLICY_ACTIONS,
                    key=lambda action: (scores[action], action == ACTION_INVOKE_HELPER),
                )
                if decision_index <= self.config.warmup_decisions:
                    selected = self._rng.choice(list(ONLINE_POLICY_ACTIONS))
                    strategy = "warmup_uniform"
                else:
                    if self._rng.random() < self.config.epsilon:
                        selected = self._rng.choice(list(ONLINE_POLICY_ACTIONS))
                        strategy = "epsilon_uniform"
                    else:
                        selected = greedy
                        strategy = "linucb_greedy"
                k_actions = float(len(ONLINE_POLICY_ACTIONS))
                base_prob = self.config.epsilon / k_actions
                if decision_index <= self.config.warmup_decisions:
                    propensity = 1.0 / k_actions
                else:
                    propensity = base_prob + (
                        (1.0 - self.config.epsilon) if selected == greedy else 0.0
                    )
                propensity = min(1.0, max(1e-9, propensity))
                applied = selected if mode == "on" else default_norm

            self._state.decisions = decision_index
            self._state.action_counts[selected] = int(self._state.action_counts.get(selected, 0)) + 1
            return PolicyDecision(
                decision_id=uuid.uuid4().hex,
                mode=mode,
                algorithm=self.config.algorithm,
                decision_index=decision_index,
                created_at_utc=_utc_now(),
                selected_action=selected,
                applied_action=applied,
                default_action=default_norm,
                propensity=float(propensity),
                decision_strategy=strategy,
                scores={key: float(value) for key, value in scores.items()},
                feature_values=cleaned_features,
            )

    def record_outcome(
        self,
        *,
        decision: PolicyDecision,
        run_id: str,
        case_id: str,
        condition: str,
        attempt: int,
        note: str,
        reward: float,
        reward_success: float,
        reward_token_penalty: float,
        reward_latency_penalty: float,
        success: bool,
        tokens_total: int,
        attempt_wall_sec: float,
        override_reason: str,
    ) -> PolicyEvent:
        with self._lock:
            model_updated = False
            update_reason = "policy_not_enabled"
            if self.config.mode == "on" and decision.applied_action in ONLINE_POLICY_ACTIONS:
                model_updated = self._update_model(
                    action=decision.applied_action,
                    reward=reward,
                    feature_values=decision.feature_values,
                )
                update_reason = (
                    "on_policy_update"
                    if model_updated
                    else "update_skipped_incomplete_state"
                )
            elif self.config.mode == "shadow":
                update_reason = "shadow_no_update"
            elif self.config.mode == "off":
                update_reason = "off_no_update"

            event = PolicyEvent(
                run_id=str(run_id),
                case_id=str(case_id),
                condition=str(condition),
                attempt=max(0, int(attempt)),
                note=str(note),
                decision_id=str(decision.decision_id),
                mode=str(decision.mode),
                algorithm=str(decision.algorithm),
                decision_index=max(0, int(decision.decision_index)),
                selected_action=str(decision.selected_action),
                applied_action=str(decision.applied_action),
                default_action=str(decision.default_action),
                propensity=min(1.0, max(1e-9, float(decision.propensity))),
                decision_strategy=str(decision.decision_strategy),
                scores={k: float(v) for k, v in decision.scores.items()},
                feature_values={k: float(v) for k, v in decision.feature_values.items()},
                reward=float(reward),
                reward_success=float(reward_success),
                reward_token_penalty=float(reward_token_penalty),
                reward_latency_penalty=float(reward_latency_penalty),
                success=bool(success),
                tokens_total=max(0, int(tokens_total)),
                attempt_wall_sec=max(0.0, float(attempt_wall_sec)),
                override_reason=str(override_reason),
                model_updated=bool(model_updated),
                update_reason=update_reason,
                created_at_utc=_utc_now(),
            )
            self._append_event(event=event)
            if model_updated:
                self._save_counter += 1
                if self._save_counter >= max(1, int(self.config.save_every)):
                    self._save_counter = 0
                    self._save_state()
            return event

    def close(self) -> None:
        with self._lock:
            self._save_state()

    def _ensure_feature_state(self, *, cleaned_features: Mapping[str, float]) -> None:
        if self.config.algorithm != "linucb":
            return
        if not self._state.feature_names:
            self._state.feature_names = sorted(str(k) for k in cleaned_features)
            dim = len(self._state.feature_names)
            if dim <= 0:
                self._state.feature_names = ["bias"]
                dim = 1
            self._state.A = {
                action: (np.eye(dim, dtype=float) * float(self.config.l2))
                for action in ONLINE_POLICY_ACTIONS
            }
            self._state.b = {
                action: np.zeros((dim,), dtype=float) for action in ONLINE_POLICY_ACTIONS
            }
            return
        expanded_feature_names = sorted(
            set(self._state.feature_names) | {str(key) for key in cleaned_features}
        )
        if expanded_feature_names != self._state.feature_names:
            previous_feature_names = list(self._state.feature_names)
            previous_index = {name: idx for idx, name in enumerate(previous_feature_names)}
            previous_dim = len(previous_feature_names)
            new_dim = len(expanded_feature_names)
            self._state.feature_names = list(expanded_feature_names)
            for action in ONLINE_POLICY_ACTIONS:
                previous_A = self._state.A.get(action)
                previous_b = self._state.b.get(action)
                next_A = np.eye(new_dim, dtype=float) * float(self.config.l2)
                next_b = np.zeros((new_dim,), dtype=float)
                if (
                    isinstance(previous_A, np.ndarray)
                    and isinstance(previous_b, np.ndarray)
                    and previous_A.shape == (previous_dim, previous_dim)
                    and previous_b.shape == (previous_dim,)
                ):
                    for new_row, feature_row in enumerate(expanded_feature_names):
                        old_row = previous_index.get(feature_row)
                        if old_row is None:
                            continue
                        next_b[new_row] = previous_b[old_row]
                        for new_col, feature_col in enumerate(expanded_feature_names):
                            old_col = previous_index.get(feature_col)
                            if old_col is None:
                                continue
                            next_A[new_row, new_col] = previous_A[old_row, old_col]
                self._state.A[action] = next_A
                self._state.b[action] = next_b
        dim = len(self._state.feature_names)
        for action in ONLINE_POLICY_ACTIONS:
            A = self._state.A.get(action)
            b = self._state.b.get(action)
            if A is None or A.shape != (dim, dim):
                self._state.A[action] = np.eye(dim, dtype=float) * float(self.config.l2)
            if b is None or b.shape != (dim,):
                self._state.b[action] = np.zeros((dim,), dtype=float)

    def _feature_vector(self, *, cleaned_features: Mapping[str, float]) -> np.ndarray:
        feature_names = self._state.feature_names
        if not feature_names:
            feature_names = ["bias"]
        values = np.array(
            [_safe_float(cleaned_features.get(name, 0.0), default=0.0) for name in feature_names],
            dtype=float,
        )
        return values

    def _linucb_scores(self, *, x: np.ndarray) -> dict[str, float]:
        scores: dict[str, float] = {}
        if self.config.algorithm != "linucb":
            return dict.fromkeys(ONLINE_POLICY_ACTIONS, 0.0)
        for action in ONLINE_POLICY_ACTIONS:
            A = self._state.A.get(action)
            b = self._state.b.get(action)
            if A is None or b is None:
                scores[action] = 0.0
                continue
            try:
                A_inv = np.linalg.inv(A)
            except np.linalg.LinAlgError:
                A = A + (np.eye(A.shape[0], dtype=float) * float(self.config.l2))
                A_inv = np.linalg.inv(A)
                self._state.A[action] = A
            theta = A_inv @ b
            exploit = float(theta.T @ x)
            explore = float(np.sqrt(max(0.0, float(x.T @ A_inv @ x))))
            scores[action] = exploit + (float(self.config.alpha) * explore)
        return scores

    def _update_model(
        self,
        *,
        action: str,
        reward: float,
        feature_values: Mapping[str, float],
    ) -> bool:
        if self.config.algorithm != "linucb":
            return False
        if action not in ONLINE_POLICY_ACTIONS:
            return False
        if not self._state.feature_names:
            return False
        x = self._feature_vector(cleaned_features=feature_values)
        A = self._state.A.get(action)
        b = self._state.b.get(action)
        if A is None or b is None:
            return False
        A += np.outer(x, x)
        b += float(reward) * x
        self._state.A[action] = A
        self._state.b[action] = b
        self._state.updates = int(self._state.updates) + 1
        return True

    def _state_payload(self) -> dict[str, Any]:
        return {
            "schema_version": STATE_SCHEMA_VERSION,
            "mode": self.config.mode,
            "algorithm": self.config.algorithm,
            "alpha": self.config.alpha,
            "epsilon": self.config.epsilon,
            "l2": self.config.l2,
            "warmup_decisions": self.config.warmup_decisions,
            "save_every": self.config.save_every,
            "feature_names": list(self._state.feature_names),
            "decisions": int(self._state.decisions),
            "updates": int(self._state.updates),
            "action_counts": {
                action: int(self._state.action_counts.get(action, 0))
                for action in ONLINE_POLICY_ACTIONS
            },
            "A": {
                action: self._state.A.get(action, np.empty((0, 0), dtype=float)).tolist()
                for action in ONLINE_POLICY_ACTIONS
            },
            "b": {
                action: self._state.b.get(action, np.empty((0,), dtype=float)).tolist()
                for action in ONLINE_POLICY_ACTIONS
            },
            "saved_at_utc": _utc_now(),
        }

    def _save_state(self) -> None:
        path = self.config.state_path
        if path is None:
            return
        payload = self._state_payload()
        path.parent.mkdir(parents=True, exist_ok=True)
        tmp_path = path.with_suffix(path.suffix + ".tmp")
        tmp_path.write_text(json.dumps(payload, ensure_ascii=False), encoding="utf-8")
        tmp_path.replace(path)

    def _load_state_if_present(self) -> None:
        path = self.config.state_path
        if path is None or not path.exists():
            return
        try:
            raw = json.loads(path.read_text(encoding="utf-8"))
            if not isinstance(raw, dict):
                raise ValueError("state payload is not an object")
            if int(raw.get("schema_version", 0)) != STATE_SCHEMA_VERSION:
                raise ValueError("unsupported schema version")
            algorithm = str(raw.get("algorithm") or "")
            if algorithm and algorithm != self.config.algorithm:
                raise ValueError("algorithm mismatch")
            feature_names_raw = raw.get("feature_names")
            if not isinstance(feature_names_raw, list) or not all(
                isinstance(item, str) for item in feature_names_raw
            ):
                raise ValueError("missing or invalid feature_names")
            feature_names = [str(item) for item in feature_names_raw]
            dim = len(feature_names)
            if dim <= 0:
                raise ValueError("empty feature_names")
            A_raw = raw.get("A")
            b_raw = raw.get("b")
            if not isinstance(A_raw, dict) or not isinstance(b_raw, dict):
                raise ValueError("missing matrices")
            A: dict[str, np.ndarray] = {}
            b: dict[str, np.ndarray] = {}
            for action in ONLINE_POLICY_ACTIONS:
                matrix_raw = A_raw.get(action)
                vector_raw = b_raw.get(action)
                matrix = np.array(matrix_raw, dtype=float)
                vector = np.array(vector_raw, dtype=float)
                if matrix.shape != (dim, dim):
                    raise ValueError(f"invalid matrix shape for {action}: {matrix.shape}")
                if vector.shape != (dim,):
                    raise ValueError(f"invalid vector shape for {action}: {vector.shape}")
                A[action] = matrix
                b[action] = vector
            action_counts_raw = raw.get("action_counts")
            action_counts = dict.fromkeys(ONLINE_POLICY_ACTIONS, 0)
            if isinstance(action_counts_raw, dict):
                for action in ONLINE_POLICY_ACTIONS:
                    action_counts[action] = _coerce_nonnegative_int(
                        action_counts_raw.get(action), default=0
                    )
            self._state = LinUCBState(
                feature_names=feature_names,
                decisions=_coerce_nonnegative_int(raw.get("decisions"), default=0),
                updates=_coerce_nonnegative_int(raw.get("updates"), default=0),
                action_counts=action_counts,
                A=A,
                b=b,
            )
            self.load_error = None
        except Exception as exc:  # pragma: no cover - defensive load hardening
            self.load_error = f"{type(exc).__name__}: {exc}"
            self._state = LinUCBState()

    def _append_event(self, *, event: PolicyEvent) -> None:
        path = self.config.events_path
        if path is None:
            return
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("a", encoding="utf-8") as handle:
            handle.write(json.dumps(event.to_json(), ensure_ascii=False) + "\n")



from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

from llmdebug import _json as json


def _bench_medians(path: Path) -> dict[str, float]:
    payload = json.loads(path.read_text(encoding="utf-8"))
    rows = payload.get("benchmarks", [])
    if not isinstance(rows, list):
        return {}
    out: dict[str, float] = {}
    for row in rows:
        if not isinstance(row, dict):
            continue
        name = str(row.get("name", "")).strip()
        if not name:
            continue
        stats: Any = row.get("stats", {})
        if not isinstance(stats, dict):
            continue
        try:
            out[name] = float(stats.get("median", 0.0))
        except (TypeError, ValueError):
            continue
    return out


def main(argv: list[str] | None = None) -> int:
    args = argv or sys.argv[1:]
    if len(args) != 2:
        print(
            "usage: bench_capture_compare.py <baseline.json> <current.json>",
            file=sys.stderr,
        )
        return 2

    baseline_path = Path(args[0])
    current_path = Path(args[1])

    baseline = _bench_medians(baseline_path)
    current = _bench_medians(current_path)
    names = sorted(set(baseline) | set(current))

    print("Capture overhead median comparison (seconds)")
    print(f"{'Benchmark':58} {'Baseline':>12} {'Current':>12} {'Delta':>12} {'Delta %':>10}")
    print("-" * 108)

    for name in names:
        base = baseline.get(name)
        curr = current.get(name)
        if base is None or curr is None:
            print(f"{name:58} {base!s:>12} {curr!s:>12} {'n/a':>12} {'n/a':>10}")
            continue
        delta = curr - base
        pct = (delta / base * 100.0) if base else float("inf")
        pct_text = "inf" if pct == float("inf") else f"{pct:+.2f}%"
        print(f"{name:58} {base:12.6f} {curr:12.6f} {delta:+12.6f} {pct_text:>10}")

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import shutil
from dataclasses import dataclass

from evals import run_eval
from evals.patchers.base import ContextCompactionResult, ContextRequestResult, PatchResult


@dataclass(frozen=True)
class _NoDiffPatcher:
    name: str = "no_diff"

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        return PatchResult(diff=None, meta={"reason": "no_diff"})


@dataclass(frozen=True)
class _RequestFailedNoDiffPatcher:
    name: str = "request_failed_no_diff"

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        return PatchResult(
            diff=None,
            meta={"error": "request_failed", "exc": "TimeoutError: timed out patcher"},
        )


@dataclass(frozen=True)
class _ContextLimitNoDiffPatcher:
    name: str = "context_limit_no_diff"

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        return PatchResult(
            diff=None,
            meta={
                "error": "http_error",
                "status": 400,
                "body": "Requested prompt is greater than the context length.",
            },
        )


@dataclass(frozen=True)
class _RetryableHttpNoDiffPatcher:
    name: str = "retryable_http_no_diff"

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        return PatchResult(
            diff=None,
            meta={
                "error": "http_error",
                "status": 503,
                "body": "backend temporarily unavailable",
            },
        )


@dataclass(frozen=True)
class _BadDiffPatcher:
    name: str = "bad_diff"

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        diff = (
            "--- a/buggy.py\n"
            "+++ b/buggy.py\n"
            "@@ -99,1 +99,1 @@\n"
            "-missing_line = 1\n"
            "+missing_line = 2\n"
        )
        return PatchResult(diff=diff, meta={"reason": "bad_diff"})


@dataclass(frozen=True)
class _MalformedStageAPatcher:
    name: str = "malformed_stage_a"

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt
        del context_dir
        del cwd
        return ContextRequestResult(
            request_json=None,
            raw_response="not-json",
            meta={"error": "bad_json"},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        return PatchResult(diff=None, meta={"reason": "no_diff"})


@dataclass(frozen=True)
class _TestFileDiffPatcher:
    name: str = "test_file_diff"

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        diff = (
            "--- a/test_buggy.py\n"
            "+++ b/test_buggy.py\n"
            "@@ -3,3 +3,3 @@\n"
            " \n"
            " def test_f() -> None:\n"
            "-    assert f() == 2\n"
            "+    assert f() == 1\n"
        )
        return PatchResult(diff=diff, meta={"reason": "test_file_edit"})


@dataclass
class _CompactingStageAPatcher:
    name: str = "compacting_stage_a"
    compact_calls: int = 0
    last_patch_prompt: str = ""

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt, context_dir, cwd
        return ContextRequestResult(
            request_json={
                "need_more_context": True,
                "need_snapshot": True,
                "need_compact_context": True,
                "snapshot_justification": "need frame locals to compare actual vs expected values",
                "compact_context_reason": "snapshot has too much noise",
                "requests": [{"type": "file", "path": "buggy.py"}],
            },
            raw_response='{"need_more_context":true}',
            meta={},
        )

    def compact_context(
        self,
        *,
        prompt: str,
        context_text: str,
        context_dir,
        cwd,
        target_chars: int,
    ) -> ContextCompactionResult:
        del context_dir, cwd
        self.compact_calls += 1
        assert "snapshot has too much noise" in prompt
        assert "Pytest output" in context_text
        compacted = "Root cause hints: buggy.f returns 1 while test expects 2."
        return ContextCompactionResult(
            compacted_text=compacted[:target_chars],
            raw_response='{"compacted_context":"ok"}',
            meta={"llm_call_count": 1, "usage": {"prompt_tokens": 5, "completion_tokens": 3}},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del context_dir, cwd
        self.last_patch_prompt = prompt
        return PatchResult(diff=None, meta={"reason": "no_diff"})


@dataclass
class _ContextRecoveryDiffPatcher:
    name: str = "context_recovery_diff"
    calls: int = 0

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        self.calls += 1
        if self.calls == 1:
            return PatchResult(
                diff=None,
                meta={
                    "error": "http_error",
                    "status": 400,
                    "body": "Requested prompt is greater than the context length.",
                },
            )
        return PatchResult(
            diff=(
                "diff --git a/buggy.py b/buggy.py\n"
                "--- a/buggy.py\n"
                "+++ b/buggy.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def f() -> int:\n"
                "-    return 1\n"
                "+    return 2\n"
            ),
            meta={"reason": "recovered_diff"},
        )


@dataclass
class _OpenAINoDiffRecoveryPatcher:
    name: str = "openai"
    calls: int = 0

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        self.calls += 1
        if self.calls == 1:
            return PatchResult(
                diff=None,
                meta={
                    "error": "no_json_payload",
                    "reason": "invalid_response",
                },
            )
        return PatchResult(
            diff=(
                "diff --git a/buggy.py b/buggy.py\n"
                "--- a/buggy.py\n"
                "+++ b/buggy.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def f() -> int:\n"
                "-    return 1\n"
                "+    return 2\n"
            ),
            meta={"reason": "recovered_diff"},
        )


@dataclass(frozen=True)
class _MarkerOnlyDiffPatcher:
    name: str = "marker_only_diff"

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        return PatchResult(
            diff=(
                "diff --git a/buggy.py b/buggy.py\n"
                "--- a/buggy.py\n"
                "+++ b/buggy.py\n"
                "@@ -1,2 +1,3 @@\n"
                " def f() -> int:\n"
                "     return 1\n"
                "+# llmdebug-force-attempt-1\n"
            ),
            meta={"reason": "marker_only"},
        )


def _make_case(tmp_path) -> run_eval.CaseSpec:
    case_dir = tmp_path / "case_src"
    case_dir.mkdir(parents=True, exist_ok=True)
    (case_dir / "buggy.py").write_text("def f() -> int:\n    return 1\n", encoding="utf-8")
    (case_dir / "test_buggy.py").write_text(
        "from buggy import f\n\n\ndef test_f() -> None:\n    assert f() == 2\n",
        encoding="utf-8",
    )
    return run_eval.CaseSpec(
        case_id="force_patch_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
        description="force-patch policy test",
    )


def test_runner_force_patch_attempt_applies_when_no_diff(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir"
    context_dir = tmp_path / "context"
    shutil.copytree(case.path, workdir)

    def _fake_run_pytest(cwd, python_args, condition, timeout_sec, output_format):
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(run_eval, "apply_diff", lambda cwd, diff: (True, ""))

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_NoDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run",
        include_files_context=False,
        force_patch_attempt=True,
    )

    assert len(records) == 1
    record = records[0]
    assert record["note"] == "forced_noop_patch"
    assert record["patch_meta"]["runner_forced_patch_attempt"] is True
    assert record["patch_meta"]["runner_forced_patch_target"] == "buggy.py"
    assert record["patch_meta"]["runner_forced_patch_marker_only"] is True
    assert record["patch_meta"]["runner_forced_patch_disabled_after_noop"] is True


def test_runner_no_force_patch_attempt_preserves_no_patch_proposed(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_no_force"
    context_dir = tmp_path / "context_no_force"
    shutil.copytree(case.path, workdir)

    def _fake_run_pytest(cwd, python_args, condition, timeout_sec, output_format):
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    apply_called = {"count": 0}

    def _fake_apply_diff(cwd, diff):
        apply_called["count"] += 1
        return True, ""

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(run_eval, "apply_diff", _fake_apply_diff)

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_NoDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    record = records[0]
    assert record["note"] == "no_patch_proposed"
    assert apply_called["count"] == 0


def test_runner_force_patch_attempt_skips_on_request_failure(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_request_failed_skip"
    context_dir = tmp_path / "context_request_failed_skip"
    shutil.copytree(case.path, workdir)

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda cwd, python_args, condition, timeout_sec, output_format: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    apply_called = {"count": 0}

    def _fake_apply_diff(cwd, diff):
        del cwd, diff
        apply_called["count"] += 1
        return True, ""

    monkeypatch.setattr(run_eval, "apply_diff", _fake_apply_diff)

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_RequestFailedNoDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run-request-failed-skip",
        include_files_context=False,
        force_patch_attempt=True,
    )

    assert len(records) == 1
    record = records[0]
    assert record["note"] == "patcher_request_failed"
    assert apply_called["count"] == 0
    assert record["patch_meta"]["patcher_request_failed"] is True
    assert record["patch_meta"]["patcher_request_failure_kind"] == "transport"
    assert record["patch_meta"]["runner_forced_patch_skipped"] is True
    assert record["patch_meta"]["runner_forced_patch_skip_reason"] == "transport_or_context_failure"
    assert "runner_forced_patch_attempt" not in record["patch_meta"]


def test_runner_force_patch_attempt_skips_on_context_limit_failure(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_context_limit_skip"
    context_dir = tmp_path / "context_context_limit_skip"
    shutil.copytree(case.path, workdir)

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda cwd, python_args, condition, timeout_sec, output_format: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    apply_called = {"count": 0}

    def _fake_apply_diff(cwd, diff):
        del cwd, diff
        apply_called["count"] += 1
        return True, ""

    monkeypatch.setattr(run_eval, "apply_diff", _fake_apply_diff)

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_ContextLimitNoDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run-context-limit-skip",
        include_files_context=False,
        force_patch_attempt=True,
    )

    assert len(records) == 1
    record = records[0]
    assert record["note"] == "patcher_request_failed"
    assert apply_called["count"] == 0
    assert record["patch_meta"]["patcher_request_failed"] is True
    assert record["patch_meta"]["patcher_request_failure_kind"] == "context_limit"
    assert record["patch_meta"]["runner_forced_patch_skipped"] is True
    assert record["patch_meta"]["runner_forced_patch_skip_reason"] == "transport_or_context_failure"
    assert "runner_forced_patch_attempt" not in record["patch_meta"]


def test_runner_force_patch_attempt_skips_on_retryable_http_transport_failure(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_retryable_http_skip"
    context_dir = tmp_path / "context_retryable_http_skip"
    shutil.copytree(case.path, workdir)

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda cwd, python_args, condition, timeout_sec, output_format: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    apply_called = {"count": 0}

    def _fake_apply_diff(cwd, diff):
        del cwd, diff
        apply_called["count"] += 1
        return True, ""

    monkeypatch.setattr(run_eval, "apply_diff", _fake_apply_diff)

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_RetryableHttpNoDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run-retryable-http-skip",
        include_files_context=False,
        force_patch_attempt=True,
    )

    assert len(records) == 1
    record = records[0]
    assert record["note"] == "patcher_request_failed"
    assert apply_called["count"] == 0
    assert record["patch_meta"]["patcher_request_failed"] is True
    assert record["patch_meta"]["patcher_request_failure_kind"] == "transport"
    assert record["patch_meta"]["runner_forced_patch_skipped"] is True
    assert record["patch_meta"]["runner_forced_patch_skip_reason"] == "transport_or_context_failure"
    assert "runner_forced_patch_attempt" not in record["patch_meta"]


def test_runner_force_patch_recovers_after_apply_failure(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_recovery"
    context_dir = tmp_path / "context_recovery"
    shutil.copytree(case.path, workdir)

    def _fake_run_pytest(cwd, python_args, condition, timeout_sec, output_format):
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    apply_attempts = {"count": 0}

    def _fake_apply_diff(cwd, diff):
        apply_attempts["count"] += 1
        if "llmdebug-force-attempt-1" in diff:
            return True, ""
        return False, "simulated hunk failure"

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(run_eval, "apply_diff", _fake_apply_diff)

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_BadDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run",
        include_files_context=False,
        force_patch_attempt=True,
    )

    assert len(records) == 1
    record = records[0]
    assert record["note"] == "tests_still_failing"
    assert record["patch_meta"]["runner_forced_patch_recovery"] is True
    assert record["patch_meta"]["runner_forced_patch_recovery_target"] == "buggy.py"
    assert apply_attempts["count"] == 2

    recovery_diff = (context_dir / "attempt_01" / "patch_recovery.diff").read_text(encoding="utf-8")
    assert "llmdebug-force-attempt-1" in recovery_diff


def test_runner_no_force_patch_keeps_apply_failure(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_recovery_off"
    context_dir = tmp_path / "context_recovery_off"
    shutil.copytree(case.path, workdir)

    def _fake_run_pytest(cwd, python_args, condition, timeout_sec, output_format):
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(run_eval, "apply_diff", lambda cwd, diff: (False, "simulated hunk failure"))

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_BadDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    record = records[0]
    assert record["note"] == "patch_apply_failed"
    assert "runner_forced_patch_recovery" not in record["patch_meta"]


def test_staged_context_malformed_stage_a_uses_fallback_bundle(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_stage_a_fallback"
    context_dir = tmp_path / "context_stage_a_fallback"
    shutil.copytree(case.path, workdir)

    def _fake_run_pytest(cwd, python_args, condition, timeout_sec, output_format):
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(run_eval, "apply_diff", lambda cwd, diff: (True, ""))

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_MalformedStageAPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run-stage-a-fallback",
        include_files_context=False,
        force_patch_attempt=False,
        context_protocol="staged",
    )

    assert len(records) == 1
    record = records[0]
    assert record["attempt_metadata"]["stage_a_request_success"] is False
    assert record["attempt_metadata"]["stage_a_requests_count"] >= 1
    assert record["patch_meta"]["stage_a_used_fallback"] is True


def test_patch_metrics_detect_test_file_edits(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_test_metric"
    context_dir = tmp_path / "context_test_metric"
    shutil.copytree(case.path, workdir)

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(run_eval, "apply_diff", lambda cwd, diff: (True, ""))

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_TestFileDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run-test-metric",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    record = records[0]
    patch_meta = record["patch_meta"]
    assert patch_meta["patch_touches_tests"] is True
    assert int(patch_meta["patch_test_file_count"]) == 1
    assert int(patch_meta["patch_file_count"]) >= 1
    assert "test_buggy.py" in patch_meta["patch_test_file_paths"]


def test_context_compaction_stage_a_request_replaces_raw_sections(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_context_compaction"
    context_dir = tmp_path / "context_context_compaction"
    shutil.copytree(case.path, workdir)
    patcher = _CompactingStageAPatcher()

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(run_eval, "apply_diff", lambda cwd, diff: (True, ""))
    monkeypatch.setattr(
        run_eval,
        "read_latest_snapshot",
        lambda *_args, **_kwargs: (
            {
                "exception": {"type": "AssertionError", "message": "expected 2 got 1"},
                "frames": [
                    {
                        "file_rel": "buggy.py",
                        "line": 2,
                        "function": "f",
                        "code": "return 1",
                        "locals": {},
                        "locals_meta": {},
                    }
                ],
            },
            None,
        ),
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_GATED,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-context-compaction",
        include_files_context=False,
        force_patch_attempt=False,
        context_protocol="staged",
        context_compaction_mode="auto_overflow",
        context_compaction_scope="snapshot,pytest,files",
        adaptive_gate_policy="aggressive",
    )

    assert len(records) == 1
    assert patcher.compact_calls == 1
    assert "Compacted context summary:" in patcher.last_patch_prompt
    assert "=== llmdebug Debug Snapshot ===" not in patcher.last_patch_prompt
    record = records[0]
    assert record["patch_meta"]["context_compaction_invoked"] is True
    assert record["patch_meta"]["context_compaction_success"] is True
    assert record["patch_meta"]["context_compaction_trigger"] == "stage_a_requested"
    assert "snapshot" in record["patch_meta"]["context_compaction_replaced_sections"]


def test_context_recovery_retries_once_with_compact_budget(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_context_recovery_retry"
    context_dir = tmp_path / "context_context_recovery_retry"
    shutil.copytree(case.path, workdir)
    patcher = _ContextRecoveryDiffPatcher()

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(run_eval, "apply_diff", lambda cwd, diff: (True, ""))

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-context-recovery-retry",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    assert patcher.calls == 2
    record = records[0]
    assert record["note"] == "tests_still_failing"
    assert record["patch_meta"]["context_recovery_invoked"] is True
    assert record["patch_meta"]["context_recovery_retry_success"] is True


def test_openai_no_diff_recovery_retries_once_before_forced_patch(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_openai_no_diff_recovery"
    context_dir = tmp_path / "context_openai_no_diff_recovery"
    shutil.copytree(case.path, workdir)
    patcher = _OpenAINoDiffRecoveryPatcher()

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(run_eval, "apply_diff", lambda cwd, diff: (True, ""))

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-openai-no-diff-recovery",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    assert patcher.calls == 2
    record = records[0]
    assert record["note"] == "tests_still_failing"
    assert record["patch_meta"]["no_diff_recovery_invoked"] is True
    assert record["patch_meta"]["no_diff_recovery_retry_success"] is True
    assert "runner_forced_patch_attempt" not in record["patch_meta"]


def test_marker_only_diff_is_treated_as_forced_noop(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_marker_noop"
    context_dir = tmp_path / "context_marker_noop"
    shutil.copytree(case.path, workdir)

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(run_eval, "apply_diff", lambda cwd, diff: (True, ""))

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.TRACEBACK_ONLY,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=_MarkerOnlyDiffPatcher(),
        k=1,
        output_format="json_compact",
        run_id="run-marker-noop",
        include_files_context=False,
        force_patch_attempt=True,
    )

    assert len(records) == 1
    record = records[0]
    assert record["note"] == "forced_noop_patch"
    assert record["patch_meta"]["runner_forced_patch_marker_only"] is True
    assert record["patch_meta"]["runner_forced_patch_disabled_after_noop"] is True



from __future__ import annotations

import os
import subprocess
from pathlib import Path

from llmdebug import _json as json

REPO_ROOT = Path(__file__).resolve().parents[1]
RUN_SCRIPT = REPO_ROOT / "evals" / "docker" / "run_eval_docker.sh"


def _install_docker_stub(bin_dir: Path) -> None:
    stub = bin_dir / "docker"
    stub.write_text(
        """#!/usr/bin/env python3
from llmdebug import _json as json
import os
import sys

log_path = os.environ.get("DOCKER_STUB_LOG", "")
if log_path:
    with open(log_path, "a", encoding="utf-8") as handle:
        handle.write(json.dumps({"argv": sys.argv[1:]}) + "\\n")

args = sys.argv[1:]
if len(args) >= 2 and args[0] == "image" and args[1] == "inspect":
    raise SystemExit(int(os.environ.get("DOCKER_STUB_IMAGE_INSPECT_RC", "0")))
if args and args[0] == "build":
    raise SystemExit(int(os.environ.get("DOCKER_STUB_BUILD_RC", "0")))
if args and args[0] == "run":
    is_probe = any("/repo/evals/results/.perm_probe" in item for item in args)
    code_var = "DOCKER_STUB_PROBE_RC" if is_probe else "DOCKER_STUB_RUN_RC"
    raise SystemExit(int(os.environ.get(code_var, "0")))
raise SystemExit(0)
""",
        encoding="utf-8",
    )
    stub.chmod(0o755)


def _install_uname_stub(bin_dir: Path) -> None:
    stub = bin_dir / "uname"
    stub.write_text(
        """#!/usr/bin/env bash
printf "%s\\n" "${DOCKER_STUB_UNAME:-Linux}"
""",
        encoding="utf-8",
    )
    stub.chmod(0o755)


def _arg_value(args: list[str], key: str) -> str | None:
    for index, arg in enumerate(args):
        if arg == key:
            if index + 1 < len(args):
                return args[index + 1]
            return None
        if arg.startswith(f"{key}="):
            return arg.split("=", 1)[1]
    return None


def _arg_values(args: list[str], key: str) -> list[str]:
    out: list[str] = []
    index = 0
    while index < len(args):
        arg = args[index]
        if arg == key:
            if index + 1 < len(args):
                out.append(args[index + 1])
            index += 2
            continue
        if arg.startswith(f"{key}="):
            out.append(arg.split("=", 1)[1])
        index += 1
    return out


def _run_wrapper(
    *,
    tmp_path: Path,
    args: list[str],
    uname_value: str,
    env_overrides: dict[str, str] | None = None,
) -> tuple[subprocess.CompletedProcess[str], list[list[str]]]:
    bin_dir = tmp_path / "bin"
    bin_dir.mkdir(parents=True, exist_ok=True)
    log_path = tmp_path / "docker_calls.jsonl"
    _install_docker_stub(bin_dir)
    _install_uname_stub(bin_dir)

    env = os.environ.copy()
    for key in (
        "OPENAI_MODEL",
        "OPENAI_BASE_URL",
        "OPENAI_API_KEY",
        "LLMDEBUG_EVAL_OPENAI_API_KEY",
        "Z_API",
        "DOCKER_EXTRA_ARGS",
    ):
        env.pop(key, None)

    env.update(
        {
            "PATH": f"{bin_dir}:{env.get('PATH', '')}",
            "DOCKER_STUB_LOG": str(log_path),
            "DOCKER_STUB_UNAME": uname_value,
            "DOCKER_STUB_IMAGE_INSPECT_RC": "0",
            "DOCKER_STUB_RUN_RC": "0",
            "LLMDEBUG_EVAL_DOCKER_USER": "1000:1000",
        }
    )
    if env_overrides:
        env.update(env_overrides)

    proc = subprocess.run(
        ["bash", str(RUN_SCRIPT), *args],
        cwd=REPO_ROOT,
        check=False,
        capture_output=True,
        text=True,
        env=env,
    )

    calls: list[list[str]] = []
    if log_path.exists():
        for line in log_path.read_text(encoding="utf-8").splitlines():
            line = line.strip()
            if not line:
                continue
            payload = json.loads(line)
            calls.append(list(payload["argv"]))
    return proc, calls


def _main_run_call(calls: list[list[str]]) -> list[str]:
    run_calls = [call for call in calls if call and call[0] == "run"]
    assert run_calls
    return run_calls[-1]


def _split_docker_and_eval_args(
    docker_run_call: list[str], *, image: str = "llmdebug-eval:latest"
) -> tuple[list[str], list[str]]:
    assert docker_run_call and docker_run_call[0] == "run"
    image_index = docker_run_call.index(image)
    docker_args = docker_run_call[1:image_index]
    eval_args = docker_run_call[image_index + 1 :]
    return docker_args, eval_args


def test_run_eval_docker_forwards_openai_autoswitch_flags_and_forces_local_executor(
    tmp_path: Path,
) -> None:
    proc, calls = _run_wrapper(
        tmp_path=tmp_path,
        uname_value="Linux",
        args=[
            "--patcher",
            "openai",
            "--openai-model",
            "qwen-demo",
            "--openai-base-url",
            "http://127.0.0.1:1234/v1",
            "--openai-response-mode",
            "edits",
            "--openai-response-mode-autoswitch",
            "retry_to_diff",
            "--openai-auto-context-window",
            "--openai-force-patch-attempt",
            "--executor",
            "testcontainers",
            "--container-network",
            "none",
            "--k",
            "3",
        ],
    )

    assert proc.returncode == 0, proc.stderr
    docker_args, eval_args = _split_docker_and_eval_args(_main_run_call(calls))
    assert _arg_value(docker_args, "--network") == "host"
    assert "--openai-auto-context-window" in eval_args
    assert _arg_value(eval_args, "--openai-response-mode-autoswitch") == "retry_to_diff"
    assert _arg_value(eval_args, "--openai-base-url") == "http://127.0.0.1:1234/v1"
    assert "--container-network" not in eval_args
    assert "testcontainers" not in eval_args
    assert _arg_value(eval_args, "--executor") == "local"


def test_run_eval_docker_forwards_no_openai_http2_flag(tmp_path: Path) -> None:
    proc, calls = _run_wrapper(
        tmp_path=tmp_path,
        uname_value="Linux",
        args=[
            "--patcher",
            "openai",
            "--openai-model",
            "qwen-demo",
            "--openai-base-url",
            "http://127.0.0.1:1234/v1",
            "--no-openai-http2",
            "--k",
            "1",
        ],
    )

    assert proc.returncode == 0, proc.stderr
    _, eval_args = _split_docker_and_eval_args(_main_run_call(calls))
    assert "--no-openai-http2" in eval_args


def test_run_eval_docker_translates_localhost_url_on_darwin(tmp_path: Path) -> None:
    proc, calls = _run_wrapper(
        tmp_path=tmp_path,
        uname_value="Darwin",
        args=[
            "--patcher",
            "openai",
            "--openai-model",
            "qwen-demo",
            "--openai-base-url",
            "http://localhost:1234/v1",
            "--k",
            "1",
        ],
    )

    assert proc.returncode == 0, proc.stderr
    docker_args, eval_args = _split_docker_and_eval_args(_main_run_call(calls))
    assert not _arg_values(docker_args, "--network")
    assert _arg_value(eval_args, "--openai-base-url") == "http://host.docker.internal:1234/v1"
    assert "Translating localhost URL for Docker access" in proc.stderr


def test_run_eval_docker_uses_openai_env_defaults_with_key_precedence(tmp_path: Path) -> None:
    proc, calls = _run_wrapper(
        tmp_path=tmp_path,
        uname_value="Linux",
        args=[
            "--patcher",
            "openai",
            "--k",
            "1",
        ],
        env_overrides={
            "OPENAI_MODEL": "env-model",
            "OPENAI_BASE_URL": "http://127.0.0.1:9999/v1",
            "OPENAI_API_KEY": "openai-key",
            "LLMDEBUG_EVAL_OPENAI_API_KEY": "llmdebug-key",
            "Z_API": "z-key",
        },
    )

    assert proc.returncode == 0, proc.stderr
    docker_args, eval_args = _split_docker_and_eval_args(_main_run_call(calls))
    assert _arg_value(eval_args, "--openai-model") == "env-model"
    assert _arg_value(eval_args, "--openai-base-url") == "http://127.0.0.1:9999/v1"
    assert _arg_value(eval_args, "--openai-api-key") == "z-key"
    assert _arg_value(docker_args, "--network") == "host"


def test_run_eval_docker_keeps_explicit_network_from_docker_extra_args(tmp_path: Path) -> None:
    proc, calls = _run_wrapper(
        tmp_path=tmp_path,
        uname_value="Linux",
        args=[
            "--patcher",
            "openai",
            "--openai-model",
            "qwen-demo",
            "--openai-base-url",
            "http://localhost:1234/v1",
            "--k",
            "1",
        ],
        env_overrides={"DOCKER_EXTRA_ARGS": "--network bridge"},
    )

    assert proc.returncode == 0, proc.stderr
    docker_args, _ = _split_docker_and_eval_args(_main_run_call(calls))
    assert _arg_values(docker_args, "--network") == ["bridge"]


def test_run_eval_docker_forwards_case_list_and_transport_retry_flags(tmp_path: Path) -> None:
    case_list_path = tmp_path / "failed_cases.txt"
    case_list_path.write_text("case_a\ncase_b\n", encoding="utf-8")

    proc, calls = _run_wrapper(
        tmp_path=tmp_path,
        uname_value="Linux",
        args=[
            "--patcher",
            "openai",
            "--openai-model",
            "qwen-demo",
            "--openai-base-url",
            "http://127.0.0.1:1234/v1",
            "--case-list-file",
            str(case_list_path),
            "--openai-transport-max-retries",
            "9",
            "--openai-transport-retry-backoff-base-sec",
            "1.5",
            "--openai-transport-retry-backoff-max-sec",
            "12.0",
            "--k",
            "1",
        ],
    )

    assert proc.returncode == 0, proc.stderr
    _, eval_args = _split_docker_and_eval_args(_main_run_call(calls))
    assert _arg_value(eval_args, "--case-list-file") == str(case_list_path)
    assert _arg_value(eval_args, "--openai-transport-max-retries") == "9"
    assert _arg_value(eval_args, "--openai-transport-retry-backoff-base-sec") == "1.5"
    assert _arg_value(eval_args, "--openai-transport-retry-backoff-max-sec") == "12.0"


def test_run_eval_docker_rewrites_config_path_for_container(tmp_path: Path) -> None:
    config_path = tmp_path / "run_eval.toml"
    config_path.write_text(
        """
[run_eval]
patcher = "null"
conditions = "traceback_only"
k = 1
""".strip()
        + "\n",
        encoding="utf-8",
    )

    proc, calls = _run_wrapper(
        tmp_path=tmp_path,
        uname_value="Linux",
        args=[
            "--config",
            str(config_path),
            "--run-id",
            "docker-config-path-rewrite",
        ],
    )

    assert proc.returncode == 0, proc.stderr
    _, eval_args = _split_docker_and_eval_args(_main_run_call(calls))
    rewritten_path = _arg_value(eval_args, "--config")
    assert rewritten_path is not None
    assert rewritten_path.startswith("/repo/evals/artifacts/config_mounts/")
    assert str(config_path) not in eval_args


def test_run_eval_docker_reports_missing_config_file(tmp_path: Path) -> None:
    missing_path = tmp_path / "missing.toml"
    proc, _calls = _run_wrapper(
        tmp_path=tmp_path,
        uname_value="Linux",
        args=[
            "--config",
            str(missing_path),
        ],
    )

    assert proc.returncode == 2
    assert "Config file not found" in proc.stderr



from __future__ import annotations

from evals.adaptive_online_policy import (
    ACTION_INVOKE_HELPER,
    ACTION_SKIP_HELPER,
    STATE_SCHEMA_VERSION,
    AdaptiveOnlinePolicyController,
    OnlinePolicyConfig,
)
from llmdebug import _json as json


def test_online_policy_on_updates_and_state_roundtrip(tmp_path) -> None:
    state_path = tmp_path / "policy_state.json"
    events_path = tmp_path / "policy_events.jsonl"
    config = OnlinePolicyConfig.normalize(
        mode="on",
        algorithm="linucb",
        alpha=0.2,
        epsilon=0.0,
        l2=1.0,
        warmup_decisions=0,
        save_every=1,
        state_path=state_path,
        events_path=events_path,
        random_seed=7,
    )
    controller = AdaptiveOnlinePolicyController(config)
    decision = controller.decide(feature_values={"bias": 1.0}, default_action=ACTION_INVOKE_HELPER)
    event = controller.record_outcome(
        decision=decision,
        run_id="run_online",
        case_id="case_a",
        condition="adaptive_llm_discretion",
        attempt=1,
        note="tests_still_failing",
        reward=0.7,
        reward_success=0.0,
        reward_token_penalty=0.2,
        reward_latency_penalty=0.1,
        success=False,
        tokens_total=123,
        attempt_wall_sec=1.2,
        override_reason="",
    )
    state = controller.state
    controller.close()

    assert event.model_updated is True
    assert event.update_reason == "on_policy_update"
    assert state.decisions == 1
    assert state.updates == 1
    assert state.action_counts[ACTION_INVOKE_HELPER] == 1
    assert events_path.exists()
    assert len([line for line in events_path.read_text(encoding="utf-8").splitlines() if line]) == 1

    reloaded = AdaptiveOnlinePolicyController(config)
    reloaded_state = reloaded.state
    reloaded.close()
    assert reloaded.load_error is None
    assert reloaded_state.decisions == 1
    assert reloaded_state.updates == 1


def test_online_policy_shadow_logs_but_does_not_update(tmp_path) -> None:
    state_path = tmp_path / "policy_state_shadow.json"
    events_path = tmp_path / "policy_events_shadow.jsonl"
    config = OnlinePolicyConfig.normalize(
        mode="shadow",
        algorithm="linucb",
        alpha=0.0,
        epsilon=0.0,
        l2=1.0,
        warmup_decisions=0,
        save_every=1,
        state_path=state_path,
        events_path=events_path,
        random_seed=0,
    )
    controller = AdaptiveOnlinePolicyController(config)
    decision = controller.decide(feature_values={"bias": 1.0}, default_action=ACTION_SKIP_HELPER)
    event = controller.record_outcome(
        decision=decision,
        run_id="run_shadow",
        case_id="case_b",
        condition="adaptive_llm_discretion",
        attempt=1,
        note="no_patch_proposed",
        reward=-0.1,
        reward_success=0.0,
        reward_token_penalty=0.1,
        reward_latency_penalty=0.0,
        success=False,
        tokens_total=10,
        attempt_wall_sec=0.2,
        override_reason="shadow_mode_baseline",
    )
    state = controller.state
    controller.close()

    assert decision.applied_action == ACTION_SKIP_HELPER
    assert event.model_updated is False
    assert event.update_reason == "shadow_no_update"
    assert state.decisions == 1
    assert state.updates == 0
    assert len([line for line in events_path.read_text(encoding="utf-8").splitlines() if line]) == 1


def test_online_policy_reports_load_error_for_invalid_state(tmp_path) -> None:
    state_path = tmp_path / "invalid_state.json"
    state_path.write_text(
        json.dumps({"schema_version": STATE_SCHEMA_VERSION, "algorithm": "linucb"}),
        encoding="utf-8",
    )
    config = OnlinePolicyConfig.normalize(
        mode="on",
        algorithm="linucb",
        state_path=state_path,
        events_path=tmp_path / "events.jsonl",
    )
    controller = AdaptiveOnlinePolicyController(config)
    state = controller.state
    controller.close()
    assert controller.load_error is not None
    assert state.feature_names == []
    assert state.decisions == 0


def test_online_policy_expands_loaded_state_for_new_feature_dimensions(tmp_path) -> None:
    state_path = tmp_path / "policy_state_expand.json"
    events_path = tmp_path / "policy_events_expand.jsonl"
    state_path.write_text(
        json.dumps(
            {
                "schema_version": STATE_SCHEMA_VERSION,
                "mode": "on",
                "algorithm": "linucb",
                "alpha": 0.2,
                "epsilon": 0.0,
                "l2": 1.5,
                "warmup_decisions": 0,
                "save_every": 1,
                "feature_names": ["bias"],
                "decisions": 3,
                "updates": 2,
                "action_counts": {
                    ACTION_SKIP_HELPER: 1,
                    ACTION_INVOKE_HELPER: 2,
                },
                "A": {
                    ACTION_SKIP_HELPER: [[3.0]],
                    ACTION_INVOKE_HELPER: [[4.0]],
                },
                "b": {
                    ACTION_SKIP_HELPER: [0.7],
                    ACTION_INVOKE_HELPER: [0.9],
                },
            },
            ensure_ascii=False,
        ),
        encoding="utf-8",
    )
    config = OnlinePolicyConfig.normalize(
        mode="on",
        algorithm="linucb",
        alpha=0.2,
        epsilon=0.0,
        l2=1.5,
        warmup_decisions=0,
        save_every=1,
        state_path=state_path,
        events_path=events_path,
        random_seed=0,
    )
    controller = AdaptiveOnlinePolicyController(config)
    controller.decide(
        feature_values={"bias": 1.0, "attempt_frac": 0.5},
        default_action=ACTION_INVOKE_HELPER,
    )
    state = controller.state
    controller.close()

    assert state.feature_names == ["attempt_frac", "bias"]
    for action in (ACTION_SKIP_HELPER, ACTION_INVOKE_HELPER):
        assert state.A[action].shape == (2, 2)
        assert state.b[action].shape == (2,)

    bias_idx = state.feature_names.index("bias")
    new_feature_idx = state.feature_names.index("attempt_frac")
    assert state.A[ACTION_SKIP_HELPER][bias_idx, bias_idx] == 3.0
    assert state.A[ACTION_INVOKE_HELPER][bias_idx, bias_idx] == 4.0
    assert state.b[ACTION_SKIP_HELPER][bias_idx] == 0.7
    assert state.b[ACTION_INVOKE_HELPER][bias_idx] == 0.9
    assert state.A[ACTION_SKIP_HELPER][new_feature_idx, new_feature_idx] == 1.5
    assert state.A[ACTION_INVOKE_HELPER][new_feature_idx, new_feature_idx] == 1.5
